from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, overload

from cognite.client._api_client import APIClient
from cognite.client.data_classes.agents import Agent, AgentList, AgentUpsert
from cognite.client.data_classes.agents.chat import (
    Action,
    ActionResult,
    AgentChatResponse,
    Message,
)
from cognite.client.utils._experimental import FeaturePreviewWarning
from cognite.client.utils._identifier import IdentifierSequence
from cognite.client.utils.useful_types import SequenceNotStr

if TYPE_CHECKING:
    from cognite.client import CogniteClient
    from cognite.client.config import ClientConfig


class AgentsAPI(APIClient):
    _RESOURCE_PATH = "/ai/agents"

    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: CogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)
        self._warnings = FeaturePreviewWarning(api_maturity="beta", sdk_maturity="alpha", feature_name="Agents")
        self._api_subversion = "beta"
        self._CREATE_LIMIT = 1
        self._DELETE_LIMIT = 1

    @overload
    def upsert(self, agents: AgentUpsert) -> Agent: ...

    @overload
    def upsert(self, agents: Sequence[AgentUpsert]) -> AgentList: ...

    def upsert(self, agents: AgentUpsert | Sequence[AgentUpsert]) -> Agent | AgentList:
        """`Create or update (upsert) one or more agents. <https://api-docs.cognite.com/20230101-beta/tag/Agents/operation/main_ai_agents_post/>`_

        Args:
            agents (AgentUpsert | Sequence[AgentUpsert]): Agent or list of agents to create or update.

        Returns:
            Agent | AgentList: The created or updated agent(s).

        Examples:

            Create a new agent with a query knowledge graph tool to find assets:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes.agents import (
                ...     AgentUpsert,
                ...     QueryKnowledgeGraphAgentToolUpsert,
                ...     QueryKnowledgeGraphAgentToolConfiguration,
                ...     DataModelInfo
                ... )
                >>> client = CogniteClient()
                ...
                >>> find_assets_tool = QueryKnowledgeGraphAgentToolUpsert(
                ...     name="find assets",
                ...     description="Use this tool to find assets",
                ...     configuration=QueryKnowledgeGraphAgentToolConfiguration(
                ...         data_models=[
                ...             DataModelInfo(
                ...                 space="cdf_idm",
                ...                 external_id="CogniteProcessIndustries",
                ...                 version="v1",
                ...                 view_external_ids=["CogniteAsset"],
                ...             )
                ...         ]
                ...     )
                ... )
                >>> agent = AgentUpsert(
                ...     external_id="my_agent",
                ...     name="My Agent",
                ...     tools=[find_assets_tool]
                ... )
                >>> client.agents.upsert(agents=[agent])

            Create an agent with multiple different tools:

                >>> from cognite.client.data_classes.agents import (
                ...     AgentUpsert,
                ...     QueryKnowledgeGraphAgentToolUpsert,
                ...     QueryKnowledgeGraphAgentToolConfiguration,
                ...     DataModelInfo,
                ...     SummarizeDocumentAgentToolUpsert,
                ...     AskDocumentAgentToolUpsert,
                ...     QueryTimeSeriesDatapointsAgentToolUpsert
                ... )
                ...
                >>> find_assets_tool = QueryKnowledgeGraphAgentToolUpsert(
                ...     name="find assets",
                ...     description="Use this tool to query the knowledge graph for assets",
                ...     configuration=QueryKnowledgeGraphAgentToolConfiguration(
                ...         data_models=[
                ...             DataModelInfo(
                ...                 space="cdf_idm",
                ...                 external_id="CogniteProcessIndustries",
                ...                 version="v1",
                ...                 view_external_ids=["CogniteAsset"],
                ...             )
                ...         ]
                ...     )
                ... )
                >>> find_files_tool = QueryKnowledgeGraphAgentToolUpsert(
                ...     name="find files",
                ...     description="Use this tool to query the knowledge graph for files",
                ...     configuration=QueryKnowledgeGraphAgentToolConfiguration(
                ...         data_models=[
                ...             DataModelInfo(
                ...                 space="cdf_idm",
                ...                 external_id="CogniteProcessIndustries",
                ...                 version="v1",
                ...                 view_external_ids=["CogniteFile"],
                ...             )
                ...         ]
                ...     )
                ... )
                >>> find_time_series_tool = QueryKnowledgeGraphAgentToolUpsert(
                ...     name="find time series",
                ...     description="Use this tool to query the knowledge graph for time series",
                ...     configuration=QueryKnowledgeGraphAgentToolConfiguration(
                ...         data_models=[
                ...             DataModelInfo(
                ...                 space="cdf_idm",
                ...                 external_id="CogniteProcessIndustries",
                ...                 version="v1",
                ...                 view_external_ids=["CogniteTimeSeries"],
                ...             )
                ...         ]
                ...     )
                ... )
                >>> summarize_tool = SummarizeDocumentAgentToolUpsert(
                ...     name="summarize document",
                ...     description="Use this tool to get a summary of a document"
                ... )
                >>> ask_doc_tool = AskDocumentAgentToolUpsert(
                ...     name="ask document",
                ...     description="Use this tool to ask questions about specific documents"
                ... )
                >>> ts_tool = QueryTimeSeriesDatapointsAgentToolUpsert(
                ...     name="query time series",
                ...     description="Use this tool to query time series data points"
                ... )
                >>> agent = AgentUpsert(
                ...     external_id="my_agent",
                ...     name="My agent",
                ...     description="An agent with many tools",
                ...     instructions="You are a helpful assistant that can query knowledge graphs, summarize documents, answer questions about documents, and query time series data points.",
                ...     tools=[find_assets_tool, find_files_tool, find_time_series_tool, summarize_tool, ask_doc_tool, ts_tool]
                ... )
                >>> client.agents.upsert(agents=[agent])


        """
        self._warnings.warn()
        return self._create_multiple(
            list_cls=AgentList,
            resource_cls=Agent,
            items=agents,
            input_resource_cls=AgentUpsert,
        )

    @overload
    def retrieve(self, external_ids: str, ignore_unknown_ids: bool = False) -> Agent | None: ...

    @overload
    def retrieve(self, external_ids: SequenceNotStr[str], ignore_unknown_ids: bool = False) -> AgentList: ...

    def retrieve(
        self, external_ids: str | SequenceNotStr[str], ignore_unknown_ids: bool = False
    ) -> Agent | AgentList | None:
        """`Retrieve one or more agents by external ID. <https://api-docs.cognite.com/20230101-beta/tag/Agents/operation/get_agents_by_ids_ai_agents_byids_post/>`_

        Args:
            external_ids (str | SequenceNotStr[str]): The external id of the agent(s) to retrieve.
            ignore_unknown_ids (bool): Whether to ignore unknown IDs. Defaults to False.

        Returns:
            Agent | AgentList | None: The requested agent or agent list. `None` is returned if `ignore_unknown_ids` is `True` and the external ID is not found.

        Examples:

            Retrieve an agent by external id:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.agents.retrieve(external_ids="my_agent")

            Retrieve multiple agents:

                >>> res = client.agents.retrieve(external_ids=["my_agent_1", "my_agent_2"])
        """
        self._warnings.warn()
        identifiers = IdentifierSequence.load(external_ids=external_ids)
        return self._retrieve_multiple(
            list_cls=AgentList,
            resource_cls=Agent,
            identifiers=identifiers,
            ignore_unknown_ids=ignore_unknown_ids,
        )

    def delete(self, external_ids: str | SequenceNotStr[str], ignore_unknown_ids: bool = False) -> None:
        """`Delete one or more agents. <https://api-docs.cognite.com/20230101-beta/tag/Agents/operation/agent_delete_ai_agents_delete_post/>`_

        Args:
            external_ids (str | SequenceNotStr[str]): External ID of the agent or a list of external ids.
            ignore_unknown_ids (bool): If `True`, the call will ignore unknown external IDs. Defaults to False.

        Examples:

            Delete an agent by external id:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> client.agents.delete(external_ids="my_agent")

        """
        self._warnings.warn()
        self._delete_multiple(
            identifiers=IdentifierSequence.load(external_ids=external_ids),
            wrap_ids=True,
            extra_body_fields={"ignoreUnknownIds": ignore_unknown_ids},
        )

    def list(self) -> AgentList:  # The API does not yet support limit or pagination
        """`List agents. <https://api-docs.cognite.com/20230101-beta/tag/Agents/operation/agent_list_ai_agents_get/>`_

        Returns:
            AgentList: The list of agents.

        Examples:

            List all agents:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> agent_list = client.agents.list()

        """
        self._warnings.warn()
        res = self._get(url_path=self._RESOURCE_PATH)
        return AgentList._load(res.json()["items"], cognite_client=self._cognite_client)

    def chat(
        self,
        agent_external_id: str,
        messages: Message | ActionResult | Sequence[Message | ActionResult],
        cursor: str | None = None,
        actions: Sequence[Action] | None = None,
    ) -> AgentChatResponse:
        """`Chat with an agent. <https://api-docs.cognite.com/20230101-beta/tag/Agents/operation/agent_session_ai_agents_chat_post/>`_

        Given a user query, the Atlas AI agent responds by reasoning and using the tools associated with it.
        Users can ensure conversation continuity by including the cursor from the previous response in subsequent requests.

        Args:
            agent_external_id (str): External ID that uniquely identifies the agent.
            messages (Message | ActionResult | Sequence[Message | ActionResult]): A list of one or many input messages to the agent. Can include regular messages and action results.
            cursor (str | None): The cursor to use for continuation of a conversation. Use this to
                create multi-turn conversations, as the cursor will keep track of the conversation state.
            actions (Sequence[Action] | None): A list of client-side actions that can be called by the agent.

        Returns:
            AgentChatResponse: The response from the agent.

        Examples:

            Start a simple chat with an agent:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes.agents import Message
                >>> client = CogniteClient()
                >>> response = client.agents.chat(
                ...     agent_external_id="my_agent",
                ...     messages=Message("What can you help me with?")
                ... )
                >>> print(response.text)

            Continue a conversation using the cursor:

                >>> follow_up = client.agents.chat(
                ...     agent_external_id="my_agent",
                ...     messages=Message("Tell me more about that"),
                ...     cursor=response.cursor
                ... )

            Send multiple messages at once:

                >>> response = client.agents.chat(
                ...     agent_external_id="my_agent",
                ...     messages=[
                ...         Message("Help me find the 1st stage compressor."),
                ...         Message("Once you have found it, find related time series.")
                ...     ]
                ... )

            Chat with client-side actions:

                >>> from cognite.client.data_classes.agents import ClientToolAction, ClientToolResult
                >>> add_numbers_action = ClientToolAction(
                ...     name="add",
                ...     description="Add two numbers together",
                ...     parameters={
                ...         "type": "object",
                ...         "properties": {
                ...             "a": {"type": "number", "description": "First number"},
                ...             "b": {"type": "number", "description": "Second number"},
                ...         },
                ...         "required": ["a", "b"]
                ...     }
                ... )
                >>> response = client.agents.chat(
                ...     agent_id="my_agent",
                ...     messages=Message("What is 42 plus 58?"),
                ...     actions=[add_numbers_action]
                ... )
                >>> if response.action_calls:
                ...     for call in response.action_calls:
                ...         # Execute the action
                ...         result = call.arguments["a"] + call.arguments["b"]
                ...         # Send result back
                ...         response = client.agents.chat(
                ...             agent_id="my_agent",
                ...             messages=ClientToolResult(
                ...                 action_id=call.action_id,
                ...                 content=f"The result is {result}"
                ...             ),
                ...             cursor=response.cursor,
                ...             actions=[add_numbers_action]
                ...         )
        """
        self._warnings.warn()

        # Convert single message to list
        if isinstance(messages, Message | ActionResult):
            messages = [messages]

        # Build request body
        body: dict[str, Any] = {
            "agentExternalId": agent_external_id,
            "messages": [msg.dump(camel_case=True) for msg in messages],
        }

        if cursor is not None:
            body["cursor"] = cursor

        if actions is not None:
            body["actions"] = [action.dump(camel_case=True) for action in actions]

        # Make the API call
        response = self._post(
            url_path=self._RESOURCE_PATH + "/chat",
            json=body,
        )

        return AgentChatResponse._load(response.json(), cognite_client=self._cognite_client)



from __future__ import annotations

from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, Literal, overload

from cognite.client._api.simulators.routine_revisions import SimulatorRoutineRevisionsAPI
from cognite.client._api_client import APIClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes.simulators.filters import PropertySort, SimulatorRoutinesFilter
from cognite.client.data_classes.simulators.routines import (
    SimulatorRoutine,
    SimulatorRoutineList,
    SimulatorRoutineWrite,
)
from cognite.client.data_classes.simulators.runs import (
    SimulationInputOverride,
    SimulationRun,
    SimulationRunWrite,
)
from cognite.client.utils._experimental import FeaturePreviewWarning
from cognite.client.utils._identifier import IdentifierSequence
from cognite.client.utils._validation import assert_type
from cognite.client.utils.useful_types import SequenceNotStr

if TYPE_CHECKING:
    from cognite.client import ClientConfig, CogniteClient


class SimulatorRoutinesAPI(APIClient):
    _RESOURCE_PATH = "/simulators/routines"

    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: CogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)
        self.revisions = SimulatorRoutineRevisionsAPI(config, api_version, cognite_client)
        self._warning = FeaturePreviewWarning(
            api_maturity="General Availability", sdk_maturity="alpha", feature_name="Simulators"
        )
        self._CREATE_LIMIT = 1
        self._DELETE_LIMIT = 1

    def __iter__(self) -> Iterator[SimulatorRoutine]:
        """Iterate over simulator routines

        Fetches simulator routines as they are iterated over, so you keep a limited number of simulator routines in memory.

        Returns:
            Iterator[SimulatorRoutine]: yields Simulator routines one by one.
        """
        return self()

    @overload
    def __call__(
        self,
        chunk_size: int,
        model_external_ids: Sequence[str] | None = None,
        simulator_integration_external_ids: Sequence[str] | None = None,
        sort: PropertySort | None = None,
        limit: int | None = None,
    ) -> Iterator[SimulatorRoutineList]: ...

    @overload
    def __call__(
        self,
        chunk_size: None = None,
        model_external_ids: Sequence[str] | None = None,
        simulator_integration_external_ids: Sequence[str] | None = None,
        sort: PropertySort | None = None,
        limit: int | None = None,
    ) -> Iterator[SimulatorRoutine]: ...

    def __call__(
        self,
        chunk_size: int | None = None,
        model_external_ids: Sequence[str] | None = None,
        simulator_integration_external_ids: Sequence[str] | None = None,
        sort: PropertySort | None = None,
        limit: int | None = None,
    ) -> Iterator[SimulatorRoutine] | Iterator[SimulatorRoutineList]:
        """Iterate over simulator routines

        Fetches simulator routines as they are iterated over, so you keep a limited number of simulator routines in memory.

        Args:
            chunk_size (int | None): Number of simulator routines to return in each chunk. Defaults to yielding one simulator routine a time.
            model_external_ids (Sequence[str] | None): Filter on model external ids.
            simulator_integration_external_ids (Sequence[str] | None): Filter on simulator integration external ids.
            sort (PropertySort | None): The criteria to sort by.
            limit (int | None): Maximum number of simulator routines to return. Defaults to return all items.

        Returns:
            Iterator[SimulatorRoutine] | Iterator[SimulatorRoutineList]: yields SimulatorRoutine one by one if chunk is not specified, else SimulatorRoutineList objects.
        """
        self._warning.warn()
        routines_filter = SimulatorRoutinesFilter(
            model_external_ids=model_external_ids,
            simulator_integration_external_ids=simulator_integration_external_ids,
        )
        return self._list_generator(
            list_cls=SimulatorRoutineList,
            resource_cls=SimulatorRoutine,
            method="POST",
            filter=routines_filter.dump(),
            sort=[PropertySort.load(sort).dump()] if sort else None,
            chunk_size=chunk_size,
            limit=limit,
        )

    @overload
    def create(self, routine: Sequence[SimulatorRoutineWrite]) -> SimulatorRoutineList: ...

    @overload
    def create(self, routine: SimulatorRoutineWrite) -> SimulatorRoutine: ...

    def create(
        self,
        routine: SimulatorRoutineWrite | Sequence[SimulatorRoutineWrite],
    ) -> SimulatorRoutine | SimulatorRoutineList:
        """`Create simulator routines <https://developer.cognite.com/api#tag/Simulator-Routines/operation/create_simulator_routine_simulators_routines_post>`_

        Args:
            routine (SimulatorRoutineWrite | Sequence[SimulatorRoutineWrite]): Simulator routine(s) to create.

        Returns:
            SimulatorRoutine | SimulatorRoutineList: Created simulator routine(s)

        Examples:
            Create new simulator routines:
                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes.simulators.routines import SimulatorRoutineWrite
                >>> client = CogniteClient()
                >>> routines = [
                ...     SimulatorRoutineWrite(
                ...         name="routine1",
                ...         external_id="routine_ext_id",
                ...         simulator_integration_external_id="integration_ext_id",
                ...         model_external_id="model_ext_id",
                ...     ),
                ...     SimulatorRoutineWrite(
                ...         name="routine2",
                ...         external_id="routine_ext_id_2",
                ...         simulator_integration_external_id="integration_ext_id_2",
                ...         model_external_id="model_ext_id_2",
                ...     )
                ... ]
                >>> res = client.simulators.routines.create(routines)
        """
        self._warning.warn()
        assert_type(routine, "simulator_routines", [SimulatorRoutineWrite, Sequence])

        return self._create_multiple(
            list_cls=SimulatorRoutineList,
            resource_cls=SimulatorRoutine,
            items=routine,
            input_resource_cls=SimulatorRoutineWrite,
            resource_path=self._RESOURCE_PATH,
        )

    def delete(
        self,
        ids: int | Sequence[int] | None = None,
        external_ids: str | SequenceNotStr[str] | SequenceNotStr[str] | None = None,
    ) -> None:
        """`Delete simulator routines <https://developer.cognite.com/api#tag/Simulator-Routines/operation/delete_simulator_routine_simulators_routines_delete_post>`_

        Args:
            ids (int | Sequence[int] | None): ids (or sequence of ids) for the routine(s) to delete.
            external_ids (str | SequenceNotStr[str] | SequenceNotStr[str] | None): external ids (or sequence of external ids) for the routine(s) to delete.

        Examples:
            Delete simulator routines by id or external id:
                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> client.simulators.routines.delete(ids=[1,2,3], external_ids="foo")
        """
        self._warning.warn()
        self._delete_multiple(
            identifiers=IdentifierSequence.load(ids=ids, external_ids=external_ids),
            wrap_ids=True,
        )

    def list(
        self,
        limit: int = DEFAULT_LIMIT_READ,
        model_external_ids: Sequence[str] | None = None,
        simulator_integration_external_ids: Sequence[str] | None = None,
        sort: PropertySort | None = None,
    ) -> SimulatorRoutineList:
        """`Filter simulator routines <https://developer.cognite.com/api#tag/Simulator-Routines/operation/filter_simulator_routines_simulators_routines_list_post>`_

        Retrieves a list of simulator routines that match the given criteria.

        Args:
            limit (int): Maximum number of results to return. Defaults to 25. Set to -1, float(“inf”) or None to return all items.
            model_external_ids (Sequence[str] | None): Filter on model external ids.
            simulator_integration_external_ids (Sequence[str] | None): Filter on simulator integration external ids.
            sort (PropertySort | None): The criteria to sort by.

        Returns:
            SimulatorRoutineList: List of simulator routines

        Examples:
            List simulator routines:
                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.simulators.routines.list(limit=10)

            Specify filter and sort order:
                >>> from cognite.client.data_classes.simulators.filters import PropertySort
                >>> res = client.simulators.routines.list(
                ...     simulator_integration_external_ids=["integration_ext_id"],
                ...     sort=PropertySort(
                ...         property="createdTime",
                ...         order="desc"
                ...     )
                ... )

        """
        self._warning.warn()
        routines_filter = SimulatorRoutinesFilter(
            model_external_ids=model_external_ids,
            simulator_integration_external_ids=simulator_integration_external_ids,
        )
        self._warning.warn()
        return self._list(
            limit=limit,
            method="POST",
            url_path="/simulators/routines/list",
            resource_cls=SimulatorRoutine,
            list_cls=SimulatorRoutineList,
            sort=[PropertySort.load(sort).dump()] if sort else None,
            filter=routines_filter.dump(),
        )

    @overload
    def run(
        self,
        *,
        routine_external_id: str,
        inputs: Sequence[SimulationInputOverride] | None = None,
        run_time: int | None = None,
        queue: bool | None = None,
        log_severity: Literal["Debug", "Information", "Warning", "Error"] | None = None,
        wait: bool = True,
        timeout: float = 60,
    ) -> SimulationRun: ...

    @overload
    def run(
        self,
        *,
        routine_revision_external_id: str,
        model_revision_external_id: str,
        inputs: Sequence[SimulationInputOverride] | None = None,
        run_time: int | None = None,
        queue: bool | None = None,
        log_severity: Literal["Debug", "Information", "Warning", "Error"] | None = None,
        wait: bool = True,
        timeout: float = 60,
    ) -> SimulationRun: ...

    def run(
        self,
        routine_external_id: str | None = None,
        routine_revision_external_id: str | None = None,
        model_revision_external_id: str | None = None,
        inputs: Sequence[SimulationInputOverride] | None = None,
        run_time: int | None = None,
        queue: bool | None = None,
        log_severity: Literal["Debug", "Information", "Warning", "Error"] | None = None,
        wait: bool = True,
        timeout: float = 60,
    ) -> SimulationRun:
        """`Run a simulation <https://developer.cognite.com/api#tag/Simulation-Runs/operation/run_simulation_simulators_run_post>`_

        Run a simulation for a given simulator routine. Supports two modes:
        1. By routine external ID only
        2. By routine revision external ID + model revision external ID

        Args:
            routine_external_id (str | None): External id of the simulator routine to run.
                Cannot be specified together with routine_revision_external_id and model_revision_external_id.
            routine_revision_external_id (str | None): External id of the simulator routine revision to run.
                Must be specified together with model_revision_external_id.
            model_revision_external_id (str | None): External id of the simulator model revision.
                Must be specified together with routine_revision_external_id.
            inputs (Sequence[SimulationInputOverride] | None): List of input overrides
            run_time (int | None): Run time in milliseconds. Reference timestamp used for data pre-processing and data sampling.
            queue (bool | None): Queue the simulation run when connector is down.
            log_severity (Literal['Debug', 'Information', 'Warning', 'Error'] | None): Override the minimum severity level for the simulation run logs. If not provided, the minimum severity is read from the connector logger configuration.
            wait (bool): Wait until the simulation run is finished. Defaults to True.
            timeout (float): Timeout in seconds for waiting for the simulation run to finish. Defaults to 60 seconds.

        Returns:
            SimulationRun: Created simulation run

        Examples:
            Create new simulation run using routine external ID:
                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> run = client.simulators.routines.run(
                ...     routine_external_id="routine1",
                ...     log_severity="Debug"
                ... )

            Create new simulation run using routine and model revision external IDs:
                >>> run = client.simulators.routines.run(
                ...     routine_revision_external_id="routine_revision1",
                ...     model_revision_external_id="model_revision1",
                ... )
        """
        self._warning.warn()
        run_object = SimulationRunWrite(
            routine_external_id=routine_external_id,
            routine_revision_external_id=routine_revision_external_id,
            model_revision_external_id=model_revision_external_id,
            inputs=list(inputs) if inputs is not None else None,
            run_time=run_time,
            queue=queue,
            log_severity=log_severity,
        )

        simulation_run = self._cognite_client.simulators.runs.create(run_object)

        if wait:
            simulation_run.wait(timeout)
        return simulation_run



from __future__ import annotations

import functools
import math
import platform
import warnings
from abc import ABC
from collections.abc import Hashable, Iterable, Iterator, Sequence
from inspect import isabstract
from typing import (
    TYPE_CHECKING,
    Any,
    TypeGuard,
    TypeVar,
    overload,
)
from urllib.parse import quote

from cognite.client.utils import _json
from cognite.client.utils._importing import local_import
from cognite.client.utils._text import (
    convert_all_keys_to_camel_case,
    convert_all_keys_to_snake_case,
    to_camel_case,
    to_snake_case,
)
from cognite.client.utils.useful_types import SequenceNotStr

if TYPE_CHECKING:
    from cognite.client import CogniteClient
    from cognite.client.data_classes._base import T_CogniteObject, T_CogniteResource

T = TypeVar("T")
K = TypeVar("K")
THashable = TypeVar("THashable", bound=Hashable)


def all_subclasses(base: type[T], exclude: set[type[T]] | None = None) -> set[type[T]]:
    """Recursively find all subclasses of a given class."""
    return set(base.__subclasses__()).union(s for c in base.__subclasses__() for s in all_subclasses(c)) - (
        exclude or set()
    )


def all_concrete_subclasses(base: type[T], exclude: set[type[T]] | None = None) -> set[type[T]]:
    """Recursively find all non-abstract subclasses of a given class."""
    return {cls for cls in all_subclasses(base, exclude) if ABC not in cls.__bases__ and not isabstract(cls)}


def no_op(x: T) -> T:
    return x


def is_finite(limit: Any) -> TypeGuard[int]:
    return isinstance(limit, int) and limit >= 0


def is_positive(limit: Any) -> TypeGuard[int]:
    return isinstance(limit, int) and limit > 0


def is_unlimited(limit: float | int | None) -> bool:
    return limit in {None, -1, math.inf}


@functools.lru_cache(None)
def get_accepted_params(cls: type[T_CogniteResource]) -> dict[str, str]:
    return {to_camel_case(k): k for k in vars(cls()) if not k.startswith("_")}


def load_resource_to_dict(resource: dict[str, Any] | str) -> dict[str, Any]:
    if isinstance(resource, dict):
        return resource

    if isinstance(resource, str):
        resource = load_yaml_or_json(resource)
        if isinstance(resource, dict):
            return resource

    raise TypeError(f"Resource must be json or yaml str, or dict, not {type(resource)}")


def fast_dict_load(
    cls: type[T_CogniteObject], item: dict[str, Any], cognite_client: CogniteClient | None
) -> T_CogniteObject:
    try:
        instance = cls(cognite_client=cognite_client)  # type: ignore [call-arg]
    except TypeError:
        instance = cls()
    # Note: Do not use cast(Hashable, cls) here as this is often called in a hot loop
    # Accepted: {camel_case(attribute_name): attribute_name}
    accepted = get_accepted_params(cls)
    for camel_attr, value in item.items():
        try:
            setattr(instance, accepted[camel_attr], value)
        except KeyError:
            pass
    return instance


def load_yaml_or_json(resource: str) -> Any:
    try:
        yaml = local_import("yaml")
        return yaml.safe_load(resource)
    except ImportError:
        return _json.loads(resource)


def basic_obj_dump(obj: Any, camel_case: bool) -> dict[str, Any]:
    if camel_case:
        return convert_all_keys_to_camel_case(vars(obj))
    return convert_all_keys_to_snake_case(vars(obj))


def handle_renamed_argument(
    new_arg: T,
    new_arg_name: str,
    old_arg_name: str,
    fn_name: str,
    kw_dct: dict[str, Any],
    required: bool = True,
) -> T:
    old_arg = kw_dct.pop(old_arg_name, None)
    if kw_dct:
        raise TypeError(f"Got unexpected keyword argument(s): {list(kw_dct)}")

    if old_arg is None:
        if new_arg is None and required:
            raise TypeError(f"{fn_name}() missing 1 required positional argument: {new_arg_name!r}")
        return new_arg

    warnings.warn(
        f"Argument {old_arg_name!r} have been changed to {new_arg_name!r}, but the old is still supported until "
        "the next major version. Consider updating your code.",
        UserWarning,
        stacklevel=2,
    )
    if new_arg is not None:
        raise TypeError(f"Pass either {new_arg_name!r} or {old_arg_name!r} (deprecated), not both")
    return old_arg


def handle_deprecated_camel_case_argument(new_arg: T, old_arg_name: str, fn_name: str, kw_dct: dict[str, Any]) -> T:
    new_arg_name = to_snake_case(old_arg_name)
    return handle_renamed_argument(new_arg, new_arg_name, old_arg_name, fn_name, kw_dct)


def interpolate_and_url_encode(path: str, *args: Any) -> str:
    return path.format(*[quote(str(arg), safe="") for arg in args])


def get_current_sdk_version() -> str:
    from cognite.client import __version__

    return __version__


@functools.lru_cache(maxsize=1)
def get_user_agent() -> str:
    sdk_version = f"CognitePythonSDK/{get_current_sdk_version()}"
    python_version = (
        f"{platform.python_implementation()}/{platform.python_version()} "
        f"({platform.python_build()};{platform.python_compiler()})"
    )
    os_version_info = [platform.release(), platform.machine(), platform.architecture()[0]]
    os_version_info = [s for s in os_version_info if s]  # Ignore empty strings
    os_version_info_str = "-".join(os_version_info)
    operating_system = f"{platform.system()}/{os_version_info_str}"

    return f"{sdk_version} {python_version} {operating_system}"


@overload
def split_into_n_parts(seq: list[T], *, n: int) -> Iterator[list[T]]: ...


@overload
def split_into_n_parts(seq: Sequence[T], *, n: int) -> Iterator[Sequence[T]]: ...


def split_into_n_parts(seq: Sequence[T], *, n: int) -> Iterator[Sequence[T]]:
    # NB: Chaotic sampling: jumps n for each starting position
    yield from (seq[i::n] for i in range(n))


@overload
def split_into_chunks(collection: set[T] | SequenceNotStr[T], chunk_size: int) -> list[list[T]]: ...


@overload
def split_into_chunks(collection: dict[K, T], chunk_size: int) -> list[dict[K, T]]: ...


def split_into_chunks(
    collection: SequenceNotStr[T] | set[T] | dict[K, T], chunk_size: int
) -> list[list[T]] | list[dict[K, T]]:
    if isinstance(collection, set):
        collection = list(collection)

    if isinstance(collection, SequenceNotStr):
        return [list(collection[i : i + chunk_size]) for i in range(0, len(collection), chunk_size)]

    if isinstance(collection, dict):
        collection = list(collection.items())
        return [dict(collection[i : i + chunk_size]) for i in range(0, len(collection), chunk_size)]

    from cognite.client.data_classes.datapoints import Datapoints, DatapointsArray

    if isinstance(collection, (DatapointsArray, Datapoints)):
        return [collection[i : i + chunk_size] for i in range(0, len(collection), chunk_size)]

    raise TypeError(f"Can only split list or dict, not {type(collection)}")


def convert_true_match(true_match: dict | list | tuple[int | str, int | str]) -> dict:
    if isinstance(true_match, Sequence) and len(true_match) == 2:
        converted_true_match = {}
        for i, fromto in enumerate(["source", "target"]):
            if isinstance(true_match[i], str):
                converted_true_match[fromto + "ExternalId"] = true_match[i]
            else:
                converted_true_match[fromto + "Id"] = true_match[i]
        return converted_true_match
    elif isinstance(true_match, dict):
        return true_match
    else:
        raise ValueError(f"true_matches should be a dictionary or a two-element list: found {true_match}")


def find_duplicates(seq: Iterable[THashable]) -> set[THashable]:
    seen: set[THashable] = set()
    add = seen.add  # skip future attr lookups for perf
    return {x for x in seq if x in seen or add(x)}


def remove_duplicates_keep_order(seq: Iterable[THashable]) -> list[THashable]:
    seen: set[THashable] = set()
    add = seen.add
    return [x for x in seq if x not in seen and not add(x)]


def exactly_one_is_not_none(*args: Any) -> bool:
    return sum(a is not None for a in args) == 1


def at_least_one_is_not_none(*args: Any) -> bool:
    return sum(a is not None for a in args) >= 1


def at_most_one_is_not_none(*args: Any) -> bool:
    return sum(a is not None for a in args) <= 1


def rename_and_exclude_keys(
    dct: dict[str, Any], aliases: dict[str, str] | None = None, exclude: set[str] | None = None
) -> dict[str, Any]:
    aliases = aliases or {}
    exclude = exclude or set()
    return {aliases.get(k, k): v for k, v in dct.items() if k not in exclude}


def load_resource(dct: dict[str, Any], cls: type[T_CogniteResource], key: str) -> T_CogniteResource | None:
    if (res := dct.get(key)) is not None:
        return cls._load(res)
    return None


def unpack_items_in_payload(payload: dict[str, dict[str, Any]]) -> list:
    return payload["json"]["items"]


def flatten_dict(d: dict[str, Any], parent_keys: tuple[str, ...], sep: str = ".") -> dict[str, Any]:
    items: list[tuple[str, Any]] = []
    for key, value in d.items():
        if isinstance(value, dict):
            items.extend(flatten_dict(value, (*parent_keys, key)).items())
        else:
            items.append((sep.join((*parent_keys, key)), value))
    return dict(items)



from __future__ import annotations

from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, Literal, overload

from cognite.client._api.workflows.executions import WorkflowExecutionAPI
from cognite.client._api.workflows.tasks import WorkflowTaskAPI
from cognite.client._api.workflows.triggers import WorkflowTriggerAPI
from cognite.client._api.workflows.versions import WorkflowVersionAPI
from cognite.client._api_client import APIClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes.workflows import (
    Workflow,
    WorkflowList,
    WorkflowUpsert,
)
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils._auxiliary import interpolate_and_url_encode, split_into_chunks
from cognite.client.utils._concurrency import execute_tasks
from cognite.client.utils._identifier import IdentifierSequence
from cognite.client.utils._validation import assert_type
from cognite.client.utils.useful_types import SequenceNotStr

if TYPE_CHECKING:
    from cognite.client import ClientConfig, CogniteClient


class WorkflowAPI(APIClient):
    _RESOURCE_PATH = "/workflows"

    def __init__(
        self,
        config: ClientConfig,
        api_version: str | None,
        cognite_client: CogniteClient,
    ) -> None:
        super().__init__(config, api_version, cognite_client)
        self.versions = WorkflowVersionAPI(config, api_version, cognite_client)
        self.executions = WorkflowExecutionAPI(config, api_version, cognite_client)
        self.tasks = WorkflowTaskAPI(config, api_version, cognite_client)
        self.triggers = WorkflowTriggerAPI(config, api_version, cognite_client)
        self._RETRIEVE_LIMIT = 1
        self._CREATE_LIMIT = 1
        self._DELETE_LIMIT = 100

    @overload
    def __call__(self, chunk_size: None = None, limit: None = None) -> Iterator[Workflow]: ...

    @overload
    def __call__(self, chunk_size: int, limit: None) -> Iterator[Workflow]: ...

    def __call__(
        self, chunk_size: int | None = None, limit: int | None = None
    ) -> Iterator[Workflow] | Iterator[WorkflowList]:
        """Iterate over workflows

        Args:
            chunk_size (int | None): The number of workflows to return in each chunk. Defaults to yielding one workflow at a time.
            limit (int | None): Maximum number of workflows to return. Defaults to returning all items.

        Returns:
            Iterator[Workflow] | Iterator[WorkflowList]: Yields Workflow one by one if chunk_size is None, otherwise yields WorkflowList objects.

        """
        return self._list_generator(
            method="GET", resource_cls=Workflow, list_cls=WorkflowList, limit=limit, chunk_size=chunk_size
        )

    def __iter__(self) -> Iterator[Workflow]:
        """Iterate all over workflows"""
        return self()

    @overload
    def upsert(self, workflow: WorkflowUpsert, mode: Literal["replace"] = "replace") -> Workflow: ...

    @overload
    def upsert(self, workflow: Sequence[WorkflowUpsert], mode: Literal["replace"] = "replace") -> WorkflowList: ...

    def upsert(
        self, workflow: WorkflowUpsert | Sequence[WorkflowUpsert], mode: Literal["replace"] = "replace"
    ) -> Workflow | WorkflowList:
        """`Create one or more workflow(s). <https://api-docs.cognite.com/20230101/tag/Workflow-versions/operation/CreateOrUpdateWorkflow>`_

        Note this is an upsert endpoint, so workflows that already exist will be updated, and new ones will be created.

        Args:
            workflow (WorkflowUpsert | Sequence[WorkflowUpsert]): The workflow(s) to upsert.
            mode (Literal['replace']): This is not an option for the API, but is included here to document that the upserts are always done in replace mode.

        Returns:
            Workflow | WorkflowList: The created workflow(s).

        Examples:

            Create one workflow with external id "my_workflow":

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import WorkflowUpsert
                >>> client = CogniteClient()
                >>> wf = WorkflowUpsert(external_id="my_workflow", description="my workflow description")
                >>> res = client.workflows.upsert(wf)

            Create multiple workflows:

                >>> wf2 = WorkflowUpsert(external_id="other", data_set_id=123)
                >>> res = client.workflows.upsert([wf, wf2])
        """
        if mode != "replace":
            raise ValueError("Only replace mode is supported for upserting workflows.")

        assert_type(workflow, "workflow", [WorkflowUpsert, Sequence])

        return self._create_multiple(
            list_cls=WorkflowList,
            resource_cls=Workflow,
            items=workflow,
            input_resource_cls=WorkflowUpsert,
        )

    @overload
    def retrieve(self, external_id: str, ignore_unknown_ids: bool = False) -> Workflow | None: ...

    @overload
    def retrieve(self, external_id: SequenceNotStr[str], ignore_unknown_ids: bool = False) -> WorkflowList: ...

    def retrieve(
        self, external_id: str | SequenceNotStr[str], ignore_unknown_ids: bool = False
    ) -> Workflow | WorkflowList | None:
        """`Retrieve one or more workflows. <https://api-docs.cognite.com/20230101/tag/Workflows/operation/fetchWorkflowDetails>`_

        Args:
            external_id (str | SequenceNotStr[str]): Identifier (or sequence of identifiers) for a Workflow. Must be unique.
            ignore_unknown_ids (bool): When requesting multiple workflows, whether to ignore external IDs that are not found rather than throwing an exception.

        Returns:
            Workflow | WorkflowList | None: If a single external ID is specified: the requested workflow, or None if it does not exist. If several external IDs are specified: the requested workflows.

        Examples:

            Retrieve workflow with external ID "my_workflow":

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> workflow = client.workflows.retrieve("my_workflow")

            Retrieve multiple workflows:

                >>> workflow_list = client.workflows.retrieve(["foo", "bar"])
        """

        # We can not use _retrieve_multiple as the backend doesn't support 'ignore_unknown_ids':
        def get_single(xid: str, ignore_missing: bool = ignore_unknown_ids) -> Workflow | None:
            try:
                response = self._get(url_path=interpolate_and_url_encode("/workflows/{}", xid))
                return Workflow._load(response.json())
            except CogniteAPIError as e:
                if ignore_missing and e.code == 404:
                    return None
                raise

        if isinstance(external_id, str):
            return get_single(external_id, ignore_missing=True)

        # Not really a point in splitting into chunks when chunk_size is 1, but...
        tasks = list(map(tuple, split_into_chunks(external_id, self._RETRIEVE_LIMIT)))
        tasks_summary = execute_tasks(get_single, tasks=tasks, max_workers=self._config.max_workers, fail_fast=True)
        tasks_summary.raise_compound_exception_if_failed_tasks()
        return WorkflowList(list(filter(None, tasks_summary.results)), cognite_client=self._cognite_client)

    def delete(self, external_id: str | SequenceNotStr[str], ignore_unknown_ids: bool = False) -> None:
        """`Delete one or more workflows with versions. <https://api-docs.cognite.com/20230101/tag/Workflows/operation/DeleteWorkflows>`_

        Args:
            external_id (str | SequenceNotStr[str]): External id or list of external ids to delete.
            ignore_unknown_ids (bool): Ignore external ids that are not found rather than throw an exception.

        Examples:

            Delete workflow with external_id "my_workflow":

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> client.workflows.delete("my_workflow")
        """
        self._delete_multiple(
            identifiers=IdentifierSequence.load(external_ids=external_id),
            params={"ignoreUnknownIds": ignore_unknown_ids},
            wrap_ids=True,
        )

    def list(self, limit: int | None = DEFAULT_LIMIT_READ) -> WorkflowList:
        """`List workflows in the project. <https://api-docs.cognite.com/20230101/tag/Workflows/operation/FetchAllWorkflows>`_

        Args:
            limit (int | None): Maximum number of results to return. Defaults to 25. Set to -1, float("inf") or None

        Returns:
            WorkflowList: Workflows in the CDF project.

        Examples:

            List all workflows:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.workflows.list(limit=None)
        """
        return self._list(
            method="GET",
            resource_cls=Workflow,
            list_cls=WorkflowList,
            limit=limit,
        )



from __future__ import annotations

from collections.abc import Sequence
from typing import overload

from cognite.client._api_client import APIClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes import (
    BoundingBox3D,
    ThreeDAssetMapping,
    ThreeDAssetMappingList,
    ThreeDAssetMappingWrite,
)
from cognite.client.utils import _json
from cognite.client.utils._auxiliary import interpolate_and_url_encode, split_into_chunks, unpack_items_in_payload
from cognite.client.utils._concurrency import execute_tasks
from cognite.client.utils._validation import assert_type


class ThreeDAssetMappingAPI(APIClient):
    _RESOURCE_PATH = "/3d/models/{}/revisions/{}/mappings"

    def list(
        self,
        model_id: int,
        revision_id: int,
        node_id: int | None = None,
        asset_id: int | None = None,
        intersects_bounding_box: BoundingBox3D | None = None,
        limit: int | None = DEFAULT_LIMIT_READ,
    ) -> ThreeDAssetMappingList:
        """`List 3D node asset mappings. <https://developer.cognite.com/api#tag/3D-Asset-Mapping/operation/get3DMappings>`_

        Args:
            model_id (int): Id of the model.
            revision_id (int): Id of the revision.
            node_id (int | None): List only asset mappings associated with this node.
            asset_id (int | None): List only asset mappings associated with this asset.
            intersects_bounding_box (BoundingBox3D | None): If given, only return asset mappings for assets whose bounding box intersects with the given bounding box.
            limit (int | None): Maximum number of asset mappings to return. Defaults to 25. Set to -1, float("inf") or None to return all items.

        Returns:
            ThreeDAssetMappingList: The list of asset mappings.

        Example:

            List 3d node asset mappings:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.three_d.asset_mappings.list(model_id=1, revision_id=1)

            List 3d node asset mappings for assets whose bounding box intersects with a given bounding box:

                >>> from cognite.client.data_classes import BoundingBox3D
                >>> bbox = BoundingBox3D(min=[0.0, 0.0, 0.0], max=[1.0, 1.0, 1.0])
                >>> res = client.three_d.asset_mappings.list(
                ...     model_id=1, revision_id=1, intersects_bounding_box=bbox)
        """
        path = interpolate_and_url_encode(self._RESOURCE_PATH, model_id, revision_id)
        flt: dict[str, str | int | None] = {"nodeId": node_id, "assetId": asset_id}
        if intersects_bounding_box:
            flt["intersectsBoundingBox"] = _json.dumps(intersects_bounding_box)
        return self._list(
            list_cls=ThreeDAssetMappingList,
            resource_cls=ThreeDAssetMapping,
            resource_path=path,
            method="GET",
            filter=flt,
            limit=limit,
        )

    @overload
    def create(
        self, model_id: int, revision_id: int, asset_mapping: ThreeDAssetMapping | ThreeDAssetMappingWrite
    ) -> ThreeDAssetMapping: ...

    @overload
    def create(
        self,
        model_id: int,
        revision_id: int,
        asset_mapping: Sequence[ThreeDAssetMapping] | Sequence[ThreeDAssetMappingWrite],
    ) -> ThreeDAssetMappingList: ...

    def create(
        self,
        model_id: int,
        revision_id: int,
        asset_mapping: ThreeDAssetMapping
        | ThreeDAssetMappingWrite
        | Sequence[ThreeDAssetMapping]
        | Sequence[ThreeDAssetMappingWrite],
    ) -> ThreeDAssetMapping | ThreeDAssetMappingList:
        """`Create 3d node asset mappings. <https://developer.cognite.com/api#tag/3D-Asset-Mapping/operation/create3DMappings>`_

        Args:
            model_id (int): Id of the model.
            revision_id (int): Id of the revision.
            asset_mapping (ThreeDAssetMapping | ThreeDAssetMappingWrite | Sequence[ThreeDAssetMapping] | Sequence[ThreeDAssetMappingWrite]): The asset mapping(s) to create.

        Returns:
            ThreeDAssetMapping | ThreeDAssetMappingList: The created asset mapping(s).

        Example:

            Create new 3d node asset mapping::

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import ThreeDAssetMappingWrite
                >>> my_mapping = ThreeDAssetMappingWrite(node_id=1, asset_id=1)
                >>> client = CogniteClient()
                >>> res = client.three_d.asset_mappings.create(model_id=1, revision_id=1, asset_mapping=my_mapping)
        """
        path = interpolate_and_url_encode(self._RESOURCE_PATH, model_id, revision_id)
        return self._create_multiple(
            list_cls=ThreeDAssetMappingList,
            resource_cls=ThreeDAssetMapping,
            resource_path=path,
            items=asset_mapping,
            input_resource_cls=ThreeDAssetMappingWrite,
        )

    def delete(
        self, model_id: int, revision_id: int, asset_mapping: ThreeDAssetMapping | Sequence[ThreeDAssetMapping]
    ) -> None:
        """`Delete 3d node asset mappings. <https://developer.cognite.com/api#tag/3D-Asset-Mapping/operation/delete3DMappings>`_

        Args:
            model_id (int): Id of the model.
            revision_id (int): Id of the revision.
            asset_mapping (ThreeDAssetMapping | Sequence[ThreeDAssetMapping]): The asset mapping(s) to delete.

        Example:

            Delete 3d node asset mapping::

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> mapping_to_delete = client.three_d.asset_mappings.list(model_id=1, revision_id=1)[0]
                >>> res = client.three_d.asset_mappings.delete(model_id=1, revision_id=1, asset_mapping=mapping_to_delete)
        """
        path = interpolate_and_url_encode(self._RESOURCE_PATH, model_id, revision_id)
        assert_type(asset_mapping, "asset_mapping", [Sequence, ThreeDAssetMapping])
        if isinstance(asset_mapping, ThreeDAssetMapping):
            asset_mapping = [asset_mapping]
        chunks = split_into_chunks(
            [ThreeDAssetMapping(a.node_id, a.asset_id).dump(camel_case=True) for a in asset_mapping], self._DELETE_LIMIT
        )
        tasks = [{"url_path": path + "/delete", "json": {"items": chunk}} for chunk in chunks]
        summary = execute_tasks(self._post, tasks, self._config.max_workers)
        summary.raise_compound_exception_if_failed_tasks(
            task_unwrap_fn=unpack_items_in_payload, task_list_element_unwrap_fn=lambda el: ThreeDAssetMapping._load(el)
        )



from __future__ import annotations

import inspect
import threading
import warnings
from abc import ABC, abstractmethod
from collections import UserDict, defaultdict
from collections.abc import (
    Collection,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    Mapping,
    MutableMapping,
    Sequence,
    ValuesView,
)
from dataclasses import dataclass
from datetime import date, datetime
from functools import lru_cache
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Literal,
    NoReturn,
    TypeAlias,
    TypeVar,
    cast,
    overload,
)

from typing_extensions import Self

from cognite.client.data_classes._base import (
    CogniteObject,
    CogniteResource,
    CogniteResourceList,
    T_CogniteResource,
    T_WriteClass,
    WriteableCogniteResourceList,
)
from cognite.client.data_classes.aggregations import AggregatedNumberedValue
from cognite.client.data_classes.data_modeling._validation import validate_data_modeling_identifier
from cognite.client.data_classes.data_modeling.core import (
    DataModelingResource,
    DataModelingSort,
    WritableDataModelingResource,
)
from cognite.client.data_classes.data_modeling.data_types import (
    DirectRelationReference,
    PropertyType,
    UnitReference,
    UnitSystemReference,
)
from cognite.client.data_classes.data_modeling.ids import (
    ContainerId,
    EdgeId,
    NodeId,
    ViewId,
    ViewIdentifier,
)
from cognite.client.utils._auxiliary import exactly_one_is_not_none, find_duplicates, flatten_dict
from cognite.client.utils._identifier import InstanceId
from cognite.client.utils._importing import local_import
from cognite.client.utils._text import convert_all_keys_to_snake_case, to_camel_case
from cognite.client.utils._time import convert_data_modelling_timestamp
from cognite.client.utils.useful_types import SequenceNotStr

if TYPE_CHECKING:
    import pandas as pd

    from cognite.client import CogniteClient
    from cognite.client.data_classes.data_modeling.debug import DebugInfo

PropertyValue: TypeAlias = (
    str | int | float | bool | dict | list[str] | list[int] | list[float] | list[bool] | list[dict]
)
PropertyValueWrite: TypeAlias = (
    str
    | int
    | float
    | bool
    | dict
    | SequenceNotStr[str]
    | Sequence[int]
    | Sequence[float]
    | Sequence[bool]
    | Sequence[dict]
    | NodeId
    | DirectRelationReference
    | date
    | datetime
    | Sequence[NodeId | DirectRelationReference]
    | Sequence[date]
    | Sequence[datetime]
    | None
)

Space: TypeAlias = str
PropertyIdentifier: TypeAlias = str


@dataclass
class NodeOrEdgeData(CogniteObject):
    """This represents the data values of a node or edge.

    Args:
        source (ContainerId | ViewId): The container or view the node or edge property is in
        properties (Mapping[str, PropertyValue]): The properties of the node or edge.
    """

    source: ContainerId | ViewId
    properties: Mapping[str, PropertyValueWrite]

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        try:
            source_type = resource["source"]["type"]
        except KeyError as e:
            raise ValueError("source must be a dict with a type key") from e
        source: ContainerId | ViewId
        if source_type == "container":
            source = ContainerId.load(resource["source"])
        elif source_type == "view":
            source = ViewId.load(resource["source"])
        else:
            raise ValueError(f"source type must be container or view, but was {source_type}")
        return cls(
            source=source,
            properties=resource["properties"],
        )

    def dump(self, camel_case: bool = True) -> dict:
        properties = _PropertyValueSerializer.serialize_values(self.properties, camel_case)
        output: dict[str, Any] = {"properties": properties}
        if self.source:
            if isinstance(self.source, (ContainerId, ViewId)):
                output["source"] = self.source.dump(camel_case)
            elif isinstance(self.source, dict):
                output["source"] = self.source
            else:
                raise TypeError(f"source must be ContainerId, ViewId or a dict, but was {type(self.source)}")
        return output


class InstanceCore(DataModelingResource, ABC):
    """A node or edge

    Args:
        space (str): The workspace for the instance, a unique identifier for the space.
        external_id (str): Combined with the space is the unique identifier of the instance.
        instance_type (Literal['node', 'edge']): The type of instance.
    """

    def __init__(self, space: str, external_id: str, instance_type: Literal["node", "edge"]) -> None:
        super().__init__(space=space)
        self.instance_type = instance_type
        self.external_id = external_id


class WritableInstanceCore(WritableDataModelingResource[T_CogniteResource], ABC):
    def __init__(self, space: str, external_id: str, instance_type: Literal["node", "edge"]) -> None:
        super().__init__(space=space)
        self.instance_type = instance_type
        self.external_id = external_id


class InstanceApply(WritableInstanceCore[T_CogniteResource], ABC):
    """A node or edge. This is the write version of the instance.

    Args:
        space (str): The workspace for the instance, a unique identifier for the space.
        external_id (str): Combined with the space is the unique identifier of the instance.
        instance_type (Literal['node', 'edge']): The type of instance.
        existing_version (int | None): Fail the ingestion request if the instance's version is greater than or equal to this value. If no existingVersion is specified, the ingestion will always overwrite any existing data for the instance (for the specified container or instance). If existingVersion is set to 0, the upsert will behave as an insert, so it will fail the bulk if the instance already exists. If skipOnVersionConflict is set on the ingestion request, then the instance will be skipped instead of failing the ingestion request.
        sources (list[NodeOrEdgeData] | None): List of source properties to write. The properties are from the instance and/or container the container(s) making up this node.
    """

    def __init__(
        self,
        space: str,
        external_id: str,
        instance_type: Literal["node", "edge"] = "node",
        existing_version: int | None = None,
        sources: list[NodeOrEdgeData] | None = None,
    ) -> None:
        validate_data_modeling_identifier(space, external_id)
        super().__init__(space, external_id, instance_type)
        self.existing_version = existing_version
        self.__sources = sources or []

    @property
    def sources(self) -> list[NodeOrEdgeData]:
        return self.__sources

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = {
            "space": self.space,
            "externalId" if camel_case else "external_id": self.external_id,
            "instanceType" if camel_case else "instance_type": self.instance_type,
        }
        if self.existing_version is not None:
            output["existingVersion" if camel_case else "existing_version"] = self.existing_version
        if self.sources:
            output["sources"] = [source.dump(camel_case) for source in self.sources]
        return output

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> Self:
        resource = convert_all_keys_to_snake_case(resource)
        if resource["instance_type"] == "node":
            return cast(Self, NodeApply._load(resource))
        if resource["instance_type"] == "edge":
            return cast(Self, EdgeApply._load(resource))
        raise ValueError(f"instance_type must be node or edge, but was {resource['instance_type']}")


_T = TypeVar("_T")


class Properties(MutableMapping[ViewIdentifier, MutableMapping[PropertyIdentifier, PropertyValue]]):
    def __init__(self, properties: MutableMapping[ViewId, MutableMapping[PropertyIdentifier, PropertyValue]]) -> None:
        self.data = properties

    @classmethod
    def load(
        cls, data: MutableMapping[Space, MutableMapping[str, MutableMapping[PropertyIdentifier, PropertyValue]]]
    ) -> Properties:
        props: MutableMapping[ViewId, MutableMapping[PropertyIdentifier, PropertyValue]] = {}
        for space, view_properties in data.items():
            for view_id_str, properties in view_properties.items():
                view_tuple = tuple(view_id_str.split("/", 1))
                if len(view_tuple) != 2:
                    warnings.warn(
                        f"Unknown type of view id: {view_id_str}, expected format <external_id>/<version>. Skipping...",
                        stacklevel=2,
                    )
                    continue
                view_id = ViewId.load((space, *view_tuple))
                props[view_id] = properties
        return cls(props)

    def dump(self) -> dict[Space, dict[str, dict[PropertyIdentifier, PropertyValue]]]:
        props: dict[Space, dict[str, dict[PropertyIdentifier, PropertyValue]]] = defaultdict(dict)
        for view_id, properties in self.data.items():
            view_id_str = f"{view_id.external_id}/{view_id.version}"
            props[view_id.space][view_id_str] = cast(dict[PropertyIdentifier, PropertyValue], properties)
        # Defaultdict is not yaml serializable
        return dict(props)

    def items(self) -> ItemsView[ViewId, MutableMapping[PropertyIdentifier, PropertyValue]]:
        return self.data.items()

    def keys(self) -> KeysView[ViewId]:
        return self.data.keys()

    def values(self) -> ValuesView[MutableMapping[PropertyIdentifier, PropertyValue]]:
        return self.data.values()

    def __iter__(self) -> Iterator[ViewId]:
        yield from self.keys()

    def __getitem__(self, view: ViewIdentifier) -> MutableMapping[PropertyIdentifier, PropertyValue]:
        view_id = ViewId.load(view)
        return self.data.get(view_id, {})

    def __contains__(self, item: Any) -> bool:
        view_id = ViewId.load(item)
        return view_id in self.data

    @overload
    def get(self, view: ViewIdentifier) -> MutableMapping[PropertyIdentifier, PropertyValue] | None: ...

    @overload
    def get(
        self, view: ViewIdentifier, default: MutableMapping[PropertyIdentifier, PropertyValue] | _T
    ) -> MutableMapping[PropertyIdentifier, PropertyValue] | _T: ...

    def get(
        self,
        view: ViewIdentifier,
        default: MutableMapping[PropertyIdentifier, PropertyValue] | None | _T | None = None,
    ) -> MutableMapping[PropertyIdentifier, PropertyValue] | None | _T:
        view_id = ViewId.load(view)
        return self.data.get(view_id, default)

    def __len__(self) -> int:
        return len(self.data)

    def __delitem__(self, view: ViewIdentifier) -> None:
        view_id = ViewId.load(view)
        del self.data[view_id]

    def __setitem__(self, view: ViewIdentifier, properties: MutableMapping[PropertyIdentifier, PropertyValue]) -> None:
        view_id = ViewId.load(view)
        self.data[view_id] = properties

    def _repr_html_(self) -> str:
        pd = local_import("pandas")
        index_names = "space", "view", "version"
        if not self:
            df = pd.DataFrame(index=pd.MultiIndex(levels=([], [], []), codes=([], [], []), names=index_names))
        else:
            df = pd.DataFrame.from_dict(
                {view_id.as_tuple(): props for view_id, props in self.data.items()},
                orient="index",
            )
            df.index.names = index_names
        return df._repr_html_()


class Instance(WritableInstanceCore[T_CogniteResource], ABC):
    """A node or edge. This is the read version of the instance.

    Args:
        space (str): The workspace for the instance, a unique identifier for the space.
        external_id (str): Combined with the space is the unique identifier of the instance.
        version (int): Current version of the instance.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        instance_type (Literal['node', 'edge']): The type of instance.
        deleted_time (int | None): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds. Timestamp when the instance was soft deleted. Note that deleted instances are filtered out of query results, but present in sync results
        properties (Properties | None): Properties of the instance.
    """

    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        instance_type: Literal["node", "edge"],
        deleted_time: int | None,
        properties: Properties | None,
    ) -> None:
        super().__init__(space, external_id, instance_type)
        self.version = version
        self.last_updated_time = last_updated_time
        self.created_time = created_time
        self.deleted_time = deleted_time
        self.__properties: Properties = properties or Properties({})

        if not isinstance(self, TypedInstance) and len(self.properties) == 1:
            (self.__prop_lookup,) = self.properties.values()
        else:
            # For speed, we want this to fail (to avoid LBYL pattern):
            self.__prop_lookup = None  # type: ignore [assignment]

    def __raise_if_non_singular_source(self, attr: str) -> NoReturn:
        if isinstance(self, TypedInstance):
            raise AttributeError(f"For typed instances, use direct attribute access: `instance.{attr}`")

        err_msg = "Quick property access is only possible on instances from a single source."
        if len(self.properties) > 1:
            err_msg += f" Hint: You may use `instance.properties[view_id][{attr!r}]`"
        raise RuntimeError(err_msg) from None

    @property
    def properties(self) -> Properties:
        return self.__properties

    @overload
    def get(self, attr: str) -> PropertyValue | None: ...

    @overload
    def get(self, attr: str, default: _T) -> PropertyValue | _T: ...

    def get(self, attr: str, default: Any = None) -> Any:
        try:
            return self.__prop_lookup.get(attr, default)
        except AttributeError:
            self.__raise_if_non_singular_source(attr)

    def __getitem__(self, attr: str) -> PropertyValue:
        try:
            return self.__prop_lookup[attr]
        except TypeError:
            self.__raise_if_non_singular_source(attr)

    def __setitem__(self, attr: str, value: PropertyValue) -> None:
        try:
            self.__prop_lookup[attr] = value
        except TypeError:
            self.__raise_if_non_singular_source(attr)

    def __delitem__(self, attr: str) -> None:
        try:
            del self.__prop_lookup[attr]
        except TypeError:
            self.__raise_if_non_singular_source(attr)

    def __contains__(self, attr: str) -> bool:
        try:
            return attr in self.__prop_lookup
        except TypeError:
            self.__raise_if_non_singular_source(attr)

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        dumped = {
            "space": self.space,
            "externalId" if camel_case else "external_id": self.external_id,
            "version": self.version,
            "lastUpdatedTime" if camel_case else "last_updated_time": self.last_updated_time,
            "createdTime" if camel_case else "created_time": self.created_time,
            "instanceType" if camel_case else "instance_type": self.instance_type,
            "properties": self.properties.dump(),
        }
        if self.deleted_time is not None:
            dumped["deletedTime" if camel_case else "deleted_time"] = self.deleted_time
        return dumped

    def to_pandas(  # type: ignore [override]
        self,
        ignore: list[str] | None = None,
        camel_case: bool = False,
        convert_timestamps: bool = True,
        expand_properties: bool = False,
        remove_property_prefix: bool = True,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """Convert the instance into a pandas DataFrame.

        Args:
            ignore (list[str] | None): List of row keys to skip when converting to a data frame. Is applied before expansions.
            camel_case (bool): Convert attribute names to camel case (e.g. `externalId` instead of `external_id`). Does not affect properties if expanded.
            convert_timestamps (bool): Convert known attributes storing CDF timestamps (milliseconds since epoch) to datetime. Does not affect properties.
            expand_properties (bool): Expand the properties into separate rows. Note: Will change default to True in the next major version.
            remove_property_prefix (bool): Attempt to remove the view ID prefix from row names of expanded properties (in index). Requires data to be from a single view and that all property names do not conflict with base properties (e.g. 'space' or 'type'). In such cases, a warning is issued and the prefix is kept.
            **kwargs (Any): For backwards compatibility.

        Returns:
            pd.DataFrame: The dataframe.
        """
        kwargs.pop("expand_metadata", None), kwargs.pop("metadata_prefix", None)
        if kwargs:
            raise TypeError(f"Unsupported keyword arguments: {kwargs}")
        if not expand_properties:
            warnings.warn(
                "Keyword argument 'expand_properties' will change default from False to True in the next major version.",
                DeprecationWarning,
            )
        df = super().to_pandas(
            expand_metadata=False, ignore=ignore, camel_case=camel_case, convert_timestamps=convert_timestamps
        )
        if not expand_properties or "properties" not in df.index:
            return df

        pd = local_import("pandas")
        col = df.squeeze()
        prop_df = pd.json_normalize(col.pop("properties"), max_level=2)
        if remove_property_prefix and not prop_df.empty:
            view_id, *extra = self.properties.keys()
            # We only do/allow this if we have a single source:
            if not extra:
                prefix = "{}.{}/{}.".format(*view_id.as_tuple())
                prop_df.columns = prop_df.columns.str.removeprefix(prefix)

                if isinstance(self, TypedInstance):
                    attr_name_mapping = self._get_descriptor_property_name_mapping()
                    prop_df = prop_df.rename(columns=attr_name_mapping)

                if any(overlapping := col.index.intersection(prop_df.columns)):
                    warnings.warn(
                        "One or more expanded property names overlapped with base properties. "
                        f"These rows (index) will not have their view ID prefix removed: {sorted(overlapping)}",
                        RuntimeWarning,
                    )
                    prop_df = prop_df.rename(columns={col: f"{prefix}{col}" for col in overlapping})
            else:
                warnings.warn(
                    "Can't remove view ID prefix from expanded property rows as source was not unique",
                    RuntimeWarning,
                )
        return pd.concat((col, prop_df.T.squeeze(axis=1))).to_frame(name="value")

    @abstractmethod
    def as_apply(self) -> InstanceApply:
        """Convert the instance to an apply instance."""
        raise NotImplementedError


class InstanceApplyResult(InstanceCore, ABC):
    """A node or edge. This represents the update on the instance.

    Args:
        instance_type (Literal['node', 'edge']): The type of instance.
        space (str): The workspace for the instance, a unique identifier for the space.
        external_id (str): Combined with the space is the unique identifier of the instance.
        version (int): DMS version of the instance.
        was_modified (bool): Whether the instance was modified by the ingestion.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
    """

    def __init__(
        self,
        instance_type: Literal["node", "edge"],
        space: str,
        external_id: str,
        version: int,
        was_modified: bool,
        last_updated_time: int,
        created_time: int,
    ) -> None:
        super().__init__(space, external_id, instance_type)
        self.version = version
        self.was_modified = was_modified
        self.last_updated_time = last_updated_time
        self.created_time = created_time


class InstanceAggregationResult(DataModelingResource):
    """Represents instances aggregation results.

    Args:
        aggregates (list[AggregatedNumberedValue]): List of aggregated values.
        group (dict[str, str | int | float | bool]): The grouping used for the aggregation.
    """

    def __init__(self, aggregates: list[AggregatedNumberedValue], group: dict[str, str | int | float | bool]) -> None:
        self.aggregates = aggregates
        self.group = group

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> Self:
        """
        Loads an instance aggregation result from a json string or dictionary.

        Args:
            resource (dict): No description.
            cognite_client (CogniteClient | None): No description.

        Returns:
            Self: An instance aggregation result.
        """
        return cls(
            aggregates=[AggregatedNumberedValue.load(agg) for agg in resource["aggregates"]],
            group=cast(dict[str, str | int | float | bool], resource.get("group")),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        """
        Dumps the aggregation results to a dictionary.

        Args:
            camel_case (bool): Whether to convert the keys to camel case.

        Returns:
            dict[str, Any]: A dictionary with the instance results.

        """
        return {
            "aggregates": [agg.dump(camel_case) for agg in self.aggregates],
            "group": self.group,
        }


class InstanceAggregationResultList(CogniteResourceList[InstanceAggregationResult]):
    _RESOURCE = InstanceAggregationResult


class InspectionResults(CogniteResource):
    def __init__(self, involved_views: list[ViewId] | None, involved_containers: list[ContainerId] | None) -> None:
        self.involved_views = involved_views
        self.involved_containers = involved_containers

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> Self:
        involved_views = (
            [ViewId.load(vid) for vid in resource["involvedViews"]] if "involvedViews" in resource else None
        )
        involved_containers = (
            [ContainerId.load(cid) for cid in resource["involvedContainers"]]
            if "involvedContainers" in resource
            else None
        )
        return cls(
            involved_views=involved_views,
            involved_containers=involved_containers,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        results = {}
        if self.involved_views:
            results["involvedViews" if camel_case else "involved_views"] = [
                vid.dump(camel_case, include_type=True) for vid in self.involved_views
            ]
        if self.involved_containers:
            results["involvedContainers" if camel_case else "involved_containers"] = [
                cid.dump(camel_case, include_type=True) for cid in self.involved_containers
            ]
        return results


@dataclass
class InspectOperation(ABC): ...


@dataclass
class InvolvedContainers(InspectOperation): ...


@dataclass
class InvolvedViews(InspectOperation):
    all_versions: bool = False


class InstanceInspectResult(CogniteResource):
    def __init__(
        self,
        space: str,
        external_id: str,
        instance_type: Literal["node", "edge"],
        inspection_results: InspectionResults,
    ) -> None:
        self.space = space
        self.external_id = external_id
        self.instance_type = instance_type
        self.inspection_results = inspection_results

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            space=resource["space"],
            external_id=resource["externalId"],
            instance_type=resource["instanceType"],
            inspection_results=InspectionResults.load(resource["inspectionResults"]),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {
            "space": self.space,
            "externalId" if camel_case else "external_id": self.external_id,
            "instanceType" if camel_case else "instance_type": self.instance_type,
            "inspectionResults" if camel_case else "inspection_results": self.inspection_results.dump(camel_case),
        }


class InstanceInspectResultList(CogniteResourceList[InstanceInspectResult]):
    _RESOURCE = InstanceInspectResult


@dataclass
class InstanceInspectResults:
    nodes: InstanceInspectResultList
    edges: InstanceInspectResultList


class NodeApply(InstanceApply["NodeApply"]):
    """A node. This is the write version of the node.

    Args:
        space (str): The workspace for the node, a unique identifier for the space.
        external_id (str): Combined with the space is the unique identifier of the node.
        existing_version (int | None): Fail the ingestion request if the node's version is greater than or equal to this value. If no existingVersion is specified, the ingestion will always overwrite any existing data for the edge (for the specified container or node). If existingVersion is set to 0, the upsert will behave as an insert, so it will fail the bulk if the item already exists. If skipOnVersionConflict is set on the ingestion request, then the item will be skipped instead of failing the ingestion request.
        sources (list[NodeOrEdgeData] | None): List of source properties to write. The properties are from the node and/or container the container(s) making up this node.
        type (DirectRelationReference | tuple[str, str] | None): Direct relation pointing to the type node.
    """

    def __init__(
        self,
        space: str,
        external_id: str,
        existing_version: int | None = None,
        sources: list[NodeOrEdgeData] | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
    ) -> None:
        super().__init__(space, external_id, "node", existing_version, sources)
        self.type = DirectRelationReference.load(type) if type else None

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> NodeApply:
        return cls(
            space=resource["space"],
            external_id=resource["externalId"],
            existing_version=resource.get("existingVersion"),
            sources=[NodeOrEdgeData.load(source) for source in resource.get("sources", [])] or None,
            type=DirectRelationReference.load(resource["type"]) if "type" in resource else None,
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if self.type:
            output["type"] = self.type.dump(camel_case)
        return output

    def as_id(self) -> NodeId:
        return NodeId(space=self.space, external_id=self.external_id)

    def as_write(self) -> Self:
        """Returns this NodeApply instance"""
        return self


class Node(Instance[NodeApply]):
    """A node. This is the read version of the node.

    Args:
        space (str): The workspace for the node, a unique identifier for the space.
        external_id (str): Combined with the space is the unique identifier of the node.
        version (int): Current version of the node.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        deleted_time (int | None): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds. Timestamp when the instance was soft deleted. Note that deleted instances are filtered out of query results, but present in sync results
        properties (Properties | None): Properties of the node.
        type (DirectRelationReference | tuple[str, str] | None): Direct relation pointing to the type node.
    """

    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        deleted_time: int | None,
        properties: Properties | None,
        type: DirectRelationReference | tuple[str, str] | None,
    ) -> None:
        super().__init__(space, external_id, version, last_updated_time, created_time, "node", deleted_time, properties)
        self.type = DirectRelationReference.load(type) if type else None

    def as_apply(self) -> NodeApply:
        """
        This is a convenience function for converting the read to a write node.

        It makes the simplifying assumption that all properties are from the same view. Note that this
        is not true in general.

        Returns:
            NodeApply: A write node, NodeApply

        """
        return NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.version,
            sources=[
                NodeOrEdgeData(source=view_id, properties=properties) for view_id, properties in self.properties.items()
            ]
            or None,
            type=self.type,
        )

    def as_write(self) -> NodeApply:
        return self.as_apply()

    def as_id(self) -> NodeId:
        return NodeId(space=self.space, external_id=self.external_id)

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if self.type:
            output["type"] = self.type.dump(camel_case)
        return output

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> Node:
        return Node(
            space=resource["space"],
            external_id=resource["externalId"],
            version=resource["version"],
            last_updated_time=resource["lastUpdatedTime"],
            created_time=resource["createdTime"],
            deleted_time=resource.get("deletedTime"),
            properties=Properties.load(resource["properties"]) if "properties" in resource else None,
            type=DirectRelationReference.load(resource["type"]) if "type" in resource else None,
        )


class NodeApplyResult(InstanceApplyResult):
    """A node. This represents the update on the node.

    Args:
        space (str): The workspace for the node, a unique identifier for the space.
        external_id (str): Combined with the space is the unique identifier of the node.
        version (int): Current version of the node.
        was_modified (bool): Whether the node was modified by the ingestion.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
    """

    def __init__(
        self, space: str, external_id: str, version: int, was_modified: bool, last_updated_time: int, created_time: int
    ) -> None:
        super().__init__(
            instance_type="node",
            space=space,
            external_id=external_id,
            version=version,
            was_modified=was_modified,
            last_updated_time=last_updated_time,
            created_time=created_time,
        )

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            space=resource["space"],
            external_id=resource["externalId"],
            version=resource["version"],
            was_modified=resource["wasModified"],
            last_updated_time=resource["lastUpdatedTime"],
            created_time=resource["createdTime"],
        )

    def as_id(self) -> NodeId:
        return NodeId(space=self.space, external_id=self.external_id)


class EdgeApply(InstanceApply["EdgeApply"]):
    """An Edge. This is the write version of the edge.

    Args:
        space (str): The workspace for the edge, a unique identifier for the space.
        external_id (str): Combined with the space is the unique identifier of the edge.
        type (DirectRelationReference | tuple[str, str]): The type of edge.
        start_node (DirectRelationReference | tuple[str, str]): Reference to the direct relation. The reference consists of a space and an external-id.
        end_node (DirectRelationReference | tuple[str, str]): Reference to the direct relation. The reference consists of a space and an external-id.
        existing_version (int | None): Fail the ingestion request if the node's version is greater than or equal to this value. If no existingVersion is specified, the ingestion will always overwrite any existing data for the edge (for the specified container or edge). If existingVersion is set to 0, the upsert will behave as an insert, so it will fail the bulk if the item already exists. If skipOnVersionConflict is set on the ingestion request, then the item will be skipped instead of failing the ingestion request.
        sources (list[NodeOrEdgeData] | None): List of source properties to write. The properties are from the edge and/or container the container(s) making up this node.
    """

    def __init__(
        self,
        space: str,
        external_id: str,
        type: DirectRelationReference | tuple[str, str],
        start_node: DirectRelationReference | tuple[str, str],
        end_node: DirectRelationReference | tuple[str, str],
        existing_version: int | None = None,
        sources: list[NodeOrEdgeData] | None = None,
    ) -> None:
        super().__init__(space, external_id, "edge", existing_version, sources)
        self.type = DirectRelationReference.load(type)
        self.start_node = DirectRelationReference.load(start_node)
        self.end_node = DirectRelationReference.load(end_node)

    def as_id(self) -> EdgeId:
        return EdgeId(space=self.space, external_id=self.external_id)

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if self.sources:
            output["sources"] = [source.dump(camel_case) for source in self.sources]
        if self.type:
            output["type"] = self.type.dump(camel_case)
        if self.start_node:
            output["startNode" if camel_case else "start_node"] = self.start_node.dump(camel_case)
        if self.end_node:
            output["endNode" if camel_case else "end_node"] = self.end_node.dump(camel_case)
        return output

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            space=resource["space"],
            external_id=resource["externalId"],
            existing_version=resource.get("existingVersion"),
            sources=[NodeOrEdgeData.load(source) for source in resource.get("sources", [])] or None,
            type=DirectRelationReference.load(resource["type"]),
            start_node=DirectRelationReference.load(resource["startNode"]),
            end_node=DirectRelationReference.load(resource["endNode"]),
        )

    def as_write(self) -> EdgeApply:
        """Returns this EdgeApply instance"""
        return self


class Edge(Instance[EdgeApply]):
    """An Edge. This is the read version of the edge.

    Args:
        space (str): The workspace for the edge, a unique identifier for the space.
        external_id (str): Combined with the space is the unique identifier of the edge.
        version (int): Current version of the edge.
        type (DirectRelationReference | tuple[str, str]): The type of edge.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        start_node (DirectRelationReference | tuple[str, str]): Reference to the direct relation. The reference consists of a space and an external-id.
        end_node (DirectRelationReference | tuple[str, str]): Reference to the direct relation. The reference consists of a space and an external-id.
        deleted_time (int | None): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds. Timestamp when the instance was soft deleted. Note that deleted instances are filtered out of query results, but present in sync results
        properties (Properties | None): No description.
    """

    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        type: DirectRelationReference | tuple[str, str],
        last_updated_time: int,
        created_time: int,
        start_node: DirectRelationReference | tuple[str, str],
        end_node: DirectRelationReference | tuple[str, str],
        deleted_time: int | None,
        properties: Properties | None,
    ) -> None:
        super().__init__(space, external_id, version, last_updated_time, created_time, "edge", deleted_time, properties)
        self.type = DirectRelationReference.load(type)
        self.start_node = DirectRelationReference.load(start_node)
        self.end_node = DirectRelationReference.load(end_node)

    def as_apply(self) -> EdgeApply:
        """
        This is a convenience function for converting the read to a write edge.

        It makes the simplifying assumption that all properties are from the same view. Note that this
        is not true in general.

        Returns:
            EdgeApply: A write edge, EdgeApply
        """
        return EdgeApply(
            space=self.space,
            external_id=self.external_id,
            type=self.type,
            start_node=self.start_node,
            end_node=self.end_node,
            existing_version=self.version,
            sources=[
                NodeOrEdgeData(source=view_id, properties=properties) for view_id, properties in self.properties.items()
            ]
            or None,
        )

    def as_write(self) -> EdgeApply:
        return self.as_apply()

    def as_id(self) -> EdgeId:
        return EdgeId(space=self.space, external_id=self.external_id)

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case)
        if self.type:
            output["type"] = self.type.dump(camel_case)
        if self.start_node:
            output["startNode" if camel_case else "start_node"] = self.start_node.dump(camel_case)
        if self.end_node:
            output["endNode" if camel_case else "end_node"] = self.end_node.dump(camel_case)
        return output

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            space=resource["space"],
            external_id=resource["externalId"],
            type=DirectRelationReference.load(resource["type"]),
            version=resource["version"],
            last_updated_time=resource["lastUpdatedTime"],
            created_time=resource["createdTime"],
            start_node=DirectRelationReference.load(resource["startNode"]),
            end_node=DirectRelationReference.load(resource["endNode"]),
            deleted_time=resource.get("deletedTime"),
            properties=Properties.load(resource["properties"]) if "properties" in resource else None,
        )


class EdgeApplyResult(InstanceApplyResult):
    """An Edge. This represents the update on the edge.

    Args:
        space (str): The workspace for the edge, a unique identifier for the space.
        external_id (str): Combined with the space is the unique identifier of the edge.
        version (int): Current version of the edge.
        was_modified (bool): Whether the edge was modified by the ingestion.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time (UTC), minus leap seconds.
    """

    def __init__(
        self, space: str, external_id: str, version: int, was_modified: bool, last_updated_time: int, created_time: int
    ) -> None:
        super().__init__(
            instance_type="edge",
            space=space,
            external_id=external_id,
            version=version,
            was_modified=was_modified,
            last_updated_time=last_updated_time,
            created_time=created_time,
        )

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            space=resource["space"],
            external_id=resource["externalId"],
            version=resource["version"],
            was_modified=resource["wasModified"],
            last_updated_time=resource["lastUpdatedTime"],
            created_time=resource["createdTime"],
        )

    def as_id(self) -> EdgeId:
        return EdgeId(space=self.space, external_id=self.external_id)


class NodeApplyResultList(CogniteResourceList[NodeApplyResult]):
    _RESOURCE = NodeApplyResult

    def as_ids(self) -> list[NodeId]:
        """
        Convert the list of nodes to a list of node ids.

        Returns:
            list[NodeId]: A list of node ids.
        """
        return [result.as_id() for result in self]


class NodeApplyList(CogniteResourceList[NodeApply]):
    _RESOURCE = NodeApply

    def as_ids(self) -> list[NodeId]:
        """
        Convert the list of nodes to a list of node ids.

        Returns:
            list[NodeId]: A list of node ids.
        """
        return [node.as_id() for node in self]


T_Instance = TypeVar("T_Instance", bound=Instance)


class DataModelingInstancesList(WriteableCogniteResourceList[T_WriteClass, T_Instance], ABC):
    def _build_id_mappings(self) -> None:
        self._instance_id_to_item = {(inst.space, inst.external_id): inst for inst in self.data}
        # We allow fast lookup via external ID when not ambiguous:
        self._ambiguous_xids = find_duplicates(inst.external_id for inst in self.data)
        self._ext_id_to_item = {inst.external_id: inst for inst in self.data}

    def get(  # type: ignore [override]
        self,
        instance_id: InstanceId | tuple[str, str] | None = None,
        external_id: str | None = None,
        *,
        id: InstanceId | tuple[str, str] | None = None,
    ) -> T_Instance | None:
        """Get an instance from this list by instance ID.

        Args:
            instance_id (InstanceId | tuple[str, str] | None): The instance ID to get. A tuple on the form (space, external_id) is also accepted.
            external_id (str | None): The external ID of the instance to return. Will raise ValueError when ambiguous (in presence of multiple spaces).
            id (InstanceId | tuple[str, str] | None): (DEPRECATED) Backwards-compatible alias for instance_id. Will be removed in the next major version.

        Returns:
            T_Instance | None: The requested instance if present, else None
        """
        if not exactly_one_is_not_none(instance_id, external_id, id):
            raise ValueError(
                "Pass exactly one of 'instance_id' or 'external_id' ('id' is a deprecated alias for 'instance_id'). "
                "Using an external ID requires all instances to be from the same space."
            )
        if id is not None:
            instance_id = id
            warnings.warn(
                "Calling .get using `id` is deprecated and will be removed in the next major version. "
                "Use 'instance_id' instead",
                UserWarning,
            )
        elif external_id is not None:
            if external_id in self._ambiguous_xids:
                raise ValueError(
                    f"{external_id=} is ambiguous (multiple spaces are present). Pass 'instance_id' instead."
                )
            return self._ext_id_to_item.get(external_id)

        if isinstance(instance_id, InstanceId):
            instance_id = instance_id.as_tuple()
        return self._instance_id_to_item.get(instance_id)

    def extend(self, other: Iterable[Any]) -> None:
        other_res_list = type(self)(other)  # See if we can accept the types
        if self._instance_id_to_item.keys().isdisjoint(other_res_list._instance_id_to_item):
            self.data.extend(other_res_list.data)
            self._build_id_mappings()
        else:
            raise ValueError("Unable to extend as this would introduce duplicates")

    def to_pandas(  # type: ignore [override]
        self,
        camel_case: bool = False,
        convert_timestamps: bool = True,
        expand_properties: bool = False,
        remove_property_prefix: bool = True,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """Convert the instance into a pandas DataFrame. Note that if the properties column is expanded and there are
        keys in the metadata that already exist in the DataFrame, then an error will be raised by pandas during joining.

        Args:
            camel_case (bool): Convert column names to camel case (e.g. `externalId` instead of `external_id`). Does not apply to properties.
            convert_timestamps (bool): Convert known columns storing CDF timestamps (milliseconds since epoch) to datetime. Does not affect properties.
            expand_properties (bool): Expand the properties into separate columns. Note: Will change default to True in the next major version.
            remove_property_prefix (bool): Attempt to remove the view ID prefix from columns names of expanded properties. Requires data to be from a single view and that all property names do not conflict with base properties (e.g. 'space' or 'type'). In such cases, a warning is issued and the prefix is kept.
            **kwargs (Any): For backwards compatibility.

        Returns:
            pd.DataFrame: The Cognite resource as a dataframe.
        """
        kwargs.pop("expand_metadata", None), kwargs.pop("metadata_prefix", None)
        if kwargs:
            raise TypeError(f"Unsupported keyword arguments: {kwargs}")
        if not expand_properties:
            warnings.warn(
                "Keyword argument 'expand_properties' will change default from False to True in the next major version.",
                DeprecationWarning,
            )
        df = super().to_pandas(camel_case=camel_case, expand_metadata=False, convert_timestamps=convert_timestamps)
        if not expand_properties or "properties" not in df.columns:
            return df

        pd = local_import("pandas")
        prop_df = pd.json_normalize(df.pop("properties"), max_level=2)
        if remove_property_prefix and not prop_df.empty:
            typed_view_ids = {item.get_source() for item in self if isinstance(item, TypedInstance)}  # type: ignore [attr-defined]
            view_id, *extra = typed_view_ids | set(vid for item in self for vid in item.properties)
            # We only do/allow this if we have a single source:
            if not extra:
                prefix = "{}.{}/{}.".format(*view_id.as_tuple())
                prop_df.columns = prop_df.columns.str.removeprefix(prefix)

                if len(self) > 0 and isinstance(self[0], TypedInstance):
                    attr_name_mapping = self[0]._get_descriptor_property_name_mapping()
                    prop_df = prop_df.rename(columns=attr_name_mapping)

                if any(overlapping := df.columns.intersection(prop_df.columns)):
                    warnings.warn(
                        "One or more expanded property names overlapped with base properties. "
                        f"These columns will not have their view ID prefix removed: {sorted(overlapping)}",
                        RuntimeWarning,
                    )
                    prop_df = prop_df.rename(columns={col: f"{prefix}{col}" for col in overlapping})
            else:
                warnings.warn(
                    "Can't remove view ID prefix from expanded property columns as multiple sources exist",
                    RuntimeWarning,
                )
        return df.join(prop_df)


T_Node = TypeVar("T_Node", bound=Node)


class NodeList(DataModelingInstancesList[NodeApply, T_Node]):
    _RESOURCE = Node  # type: ignore[assignment]

    def __init__(
        self,
        resources: Collection[Any],
        typing: TypeInformation | None = None,
        debug: DebugInfo | None = None,
        cognite_client: CogniteClient | None = None,
    ) -> None:
        super().__init__(resources, cognite_client)
        self.typing = typing
        self.debug = debug

    def as_ids(self) -> list[NodeId]:
        """
        Convert the list of nodes to a list of node ids.

        Returns:
            list[NodeId]: A list of node ids.
        """
        return [node.as_id() for node in self]

    def as_write(self) -> NodeApplyList:
        """Returns this NodeList as a NodeApplyList"""
        return NodeApplyList([node.as_write() for node in self])

    @classmethod
    def _load_raw_api_response(cls, responses: list[dict[str, Any]], cognite_client: CogniteClient) -> Self:
        from cognite.client.data_classes.data_modeling.debug import DebugInfo

        typing = next((TypeInformation._load(resp["typing"]) for resp in responses if "typing" in resp), None)
        debug = next((DebugInfo._load(r["debug"]) for r in responses if "debug" in r), None)
        resources = [
            cls._RESOURCE._load(item, cognite_client=cognite_client)
            for response in responses
            for item in response.get("items", [])
        ]
        return cls(resources, typing, debug, cognite_client=cognite_client)

    def dump_raw(self, camel_case: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = {
            "items": self.dump(camel_case),
        }
        if self.typing:
            output["typing"] = self.typing.dump(camel_case)
        if self.debug:
            output["debug"] = self.debug.dump(camel_case)
        return output


class NodeListWithCursor(NodeList[T_Node]):
    def __init__(
        self,
        resources: Collection[Any],
        cursor: str | None,
        typing: TypeInformation | None = None,
        debug: DebugInfo | None = None,
        cognite_client: CogniteClient | None = None,
    ) -> None:
        super().__init__(resources, typing, debug, cognite_client)
        self.cursor = cursor

    def extend(self, other: NodeListWithCursor) -> None:  # type: ignore[override]
        if not isinstance(other, type(self)):
            raise TypeError("Unable to extend as the types do not match")
        other_res_list = type(self)(other, other.cursor)  # See if we can accept the types
        if self._instance_id_to_item.keys().isdisjoint(other_res_list._instance_id_to_item):
            self.data.extend(other.data)
            self._build_id_mappings()
            self.cursor = other.cursor
        else:
            raise ValueError("Unable to extend as this would introduce duplicates")


class EdgeApplyResultList(CogniteResourceList[EdgeApplyResult]):
    _RESOURCE = EdgeApplyResult

    def as_ids(self) -> list[EdgeId]:
        """
        Convert the list of edges to a list of edge ids.

        Returns:
            list[EdgeId]: A list of edge ids.
        """
        return [edge.as_id() for edge in self]


class EdgeApplyList(CogniteResourceList[EdgeApply]):
    _RESOURCE = EdgeApply

    def as_ids(self) -> list[EdgeId]:
        """
        Convert the list of edges to a list of edge ids.

        Returns:
            list[EdgeId]: A list of edge ids.
        """
        return [edge.as_id() for edge in self]


T_Edge = TypeVar("T_Edge", bound=Edge)


class EdgeList(DataModelingInstancesList[EdgeApply, T_Edge]):
    _RESOURCE = Edge  # type: ignore[assignment]

    def __init__(
        self,
        resources: Collection[Any],
        typing: TypeInformation | None = None,
        debug: DebugInfo | None = None,
        cognite_client: CogniteClient | None = None,
    ) -> None:
        super().__init__(resources, cognite_client)
        self.typing = typing
        self.debug = debug

    def as_ids(self) -> list[EdgeId]:
        """
        Convert the list of edges to a list of edge ids.

        Returns:
            list[EdgeId]: A list of edge ids.
        """
        return [edge.as_id() for edge in self]

    def as_write(self) -> EdgeApplyList:
        """Returns this EdgeList as a EdgeApplyList"""
        return EdgeApplyList([edge.as_write() for edge in self], cognite_client=self._get_cognite_client())

    @classmethod
    def _load_raw_api_response(cls, responses: list[dict[str, Any]], cognite_client: CogniteClient) -> Self:
        from cognite.client.data_classes.data_modeling.debug import DebugInfo

        typing = next((TypeInformation._load(resp["typing"]) for resp in responses if "typing" in resp), None)
        debug = next((DebugInfo._load(r["debug"]) for r in responses if "debug" in r), None)
        resources = [
            cls._RESOURCE._load(item, cognite_client=cognite_client)
            for response in responses
            for item in response.get("items", [])
        ]
        return cls(resources, typing, debug, cognite_client=cognite_client)

    def dump_raw(self, camel_case: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = {
            "items": self.dump(camel_case),
        }
        if self.typing:
            output["typing"] = self.typing.dump(camel_case)
        if self.debug:
            output["debug"] = self.debug.dump(camel_case)
        return output


class EdgeListWithCursor(EdgeList):
    def __init__(
        self,
        resources: Collection[Any],
        cursor: str | None,
        typing: TypeInformation | None = None,
        debug: DebugInfo | None = None,
        cognite_client: CogniteClient | None = None,
    ) -> None:
        super().__init__(resources, typing, debug, cognite_client)
        self.cursor = cursor

    def extend(self, other: EdgeListWithCursor) -> None:  # type: ignore[override]
        if not isinstance(other, type(self)):
            raise TypeError("Unable to extend as the types do not match")
        other_res_list = type(self)(other, other.cursor)  # See if we can accept the types
        if self._instance_id_to_item.keys().isdisjoint(other_res_list._instance_id_to_item):
            self.data.extend(other.data)
            self._build_id_mappings()
            self.cursor = other.cursor
        else:
            raise ValueError("Unable to extend as this would introduce duplicates")


# This is a utility class. It is not used by in the SDK codebase, but used in projects that use the SDK.
@dataclass
class InstancesApply:
    """
    This represents the write request of an instance query

    Args:
        nodes (NodeApplyList): A list of nodes.
        edges (EdgeApplyList): A list of edges.
    """

    nodes: NodeApplyList
    edges: EdgeApplyList


class InstanceSort(DataModelingSort):
    def __init__(
        self,
        property: list[str] | tuple[str, ...],
        direction: Literal["ascending", "descending"] = "ascending",
        nulls_first: bool | None = None,
    ) -> None:
        super().__init__(property, direction, nulls_first)


@dataclass
class InstancesResult(Generic[T_Node, T_Edge]):
    """This represents the read result of an instance query

    Args:
        nodes (NodeList): A list of nodes.
        edges (EdgeList): A list of edges.

    """

    nodes: NodeList[T_Node]
    edges: EdgeList[T_Edge]


@dataclass
class InstancesApplyResult:
    """This represents the write result of an instance query

    Args:
        nodes (NodeApplyResultList): A list of nodes.
        edges (EdgeApplyResultList): A list of edges.

    """

    nodes: NodeApplyResultList
    edges: EdgeApplyResultList


@dataclass
class InstancesDeleteResult:
    """This represents the delete result of an instance query

    Args:
        nodes (list[NodeId]): A list of node ids.
        edges (list[EdgeId]): A list of edge ids.

    """

    nodes: list[NodeId]
    edges: list[EdgeId]


@dataclass
class SubscriptionContext:
    last_successful_sync: datetime | None = None
    last_successful_callback: datetime | None = None
    _canceled: bool = False
    _thread: threading.Thread | None = None

    def cancel(self) -> None:
        self._canceled = True

    def is_alive(self) -> bool:
        return self._thread is not None and self._thread.is_alive()


@dataclass
class TargetUnit(CogniteObject):
    property: str
    unit: UnitReference | UnitSystemReference

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        return {"property": self.property, "unit": self.unit.dump(camel_case)}

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> TargetUnit:
        return cls(
            property=resource["property"],
            unit=UnitReference.load(resource["unit"])
            if "externalId" in resource["unit"]
            else UnitSystemReference.load(resource["unit"]),
        )


@dataclass
class TypePropertyDefinition(CogniteObject):
    type: PropertyType
    nullable: bool = True
    auto_increment: bool = False
    immutable: bool = False
    default_value: str | int | dict | None = None
    name: str | None = None
    description: str | None = None

    def dump(self, camel_case: bool = True, return_flat_dict: bool = False) -> dict[str, Any]:
        output: dict[str, Any] = {}
        if return_flat_dict:
            dumped = flatten_dict(self.type.dump(camel_case), ("type",), sep=".")
            output.update({key: value for key, value in dumped.items()})
        else:
            output["type"] = self.type.dump(camel_case)
        output.update(
            {
                "nullable": self.nullable,
                "autoIncrement": self.auto_increment,
                "defaultValue": self.default_value,
                "name": self.name,
                "description": self.description,
                "immutable": self.immutable,
            }
        )
        return output

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> TypePropertyDefinition:
        return cls(
            type=PropertyType.load(resource["type"]),
            nullable=resource.get("nullable"),  # type: ignore[arg-type]
            immutable=resource.get("immutable"),  # type: ignore[arg-type]
            auto_increment=resource.get("autoIncrement"),  # type: ignore[arg-type]
            default_value=resource.get("defaultValue"),
            name=resource.get("name"),
            description=resource.get("description"),
        )


class TypeInformation(UserDict, CogniteObject):
    def __init__(self, data: dict[str, dict[str, dict[str, TypePropertyDefinition]]] | None = None) -> None:
        super().__init__(data or {})

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output: dict[str, Any] = {}
        for space_name, space_data in self.items():
            output.setdefault(space_name, {})
            for view_or_container_id, view_data in space_data.items():
                output[space_name].setdefault(view_or_container_id, {})
                for type_name, type_data in view_data.items():
                    output[space_name][view_or_container_id][type_name] = type_data.dump(camel_case)
        return output

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> TypeInformation:
        return cls(
            {
                space_name: {
                    view_or_container_id: {
                        type_name: TypePropertyDefinition.load(type_data)
                        for type_name, type_data in view_data.items()
                        if isinstance(type_data, dict)
                    }
                    for view_or_container_id, view_data in space_data.items()
                    if isinstance(view_data, dict)
                }
                for space_name, space_data in resource.items()
                if isinstance(space_data, dict)
            }
        )

    def to_pandas(self) -> pd.DataFrame:
        pd = local_import("pandas")

        index_names = "space_name", "view_or_container"
        if not self:
            df = pd.DataFrame(index=pd.MultiIndex(levels=([], []), codes=([], []), names=index_names))
        else:
            df = pd.DataFrame.from_dict(
                {
                    (space_name, view_or_container_id): {
                        "identifier": type_name,
                        **type_data.dump(camel_case=False, return_flat_dict=True),
                    }
                    for space_name, space_data in self.data.items()
                    for view_or_container_id, view_data in space_data.items()
                    for type_name, type_data in view_data.items()
                },
                orient="index",
            )
            df.index.names = index_names
        return df

    def _repr_html_(self) -> str:
        return self.to_pandas()._repr_html_()

    @overload
    def __getitem__(self, item: str) -> dict[str, dict[str, TypePropertyDefinition]]: ...

    @overload
    def __getitem__(self, item: tuple[str, str]) -> dict[str, TypePropertyDefinition]: ...

    @overload
    def __getitem__(self, item: tuple[str, str, str]) -> TypePropertyDefinition: ...

    def __getitem__(
        self, item: str | tuple[str, str] | tuple[str, str, str]
    ) -> dict[str, dict[str, TypePropertyDefinition]] | dict[str, TypePropertyDefinition] | TypePropertyDefinition:
        if isinstance(item, str):
            return super().__getitem__(item)
        elif isinstance(item, tuple) and len(item) == 2:
            return super().__getitem__(item[0])[item[1]]
        elif isinstance(item, tuple) and len(item) == 3:
            return super().__getitem__(item[0])[item[1]][item[2]]
        else:
            raise ValueError(f"Invalid key: {item}")

    def __setitem__(self, key: str | tuple[str, str] | tuple[str, str, str], value: Any) -> None:
        if isinstance(key, str):
            super().__setitem__(key, value)
        elif isinstance(key, tuple) and len(key) == 2:
            super().__setitem__(key[0], {key[1]: value})
        elif isinstance(key, tuple) and len(key) == 3:
            super().__setitem__(key[0], {key[1]: {key[2]: value}})
        else:
            raise ValueError(f"Invalid key: {key}")

    def __delitem__(self, key: str | tuple[str, str] | tuple[str, str, str]) -> None:
        if isinstance(key, str):
            super().__delitem__(key)
        elif isinstance(key, tuple) and len(key) == 2:
            del self[key[0]][key[1]]
        elif isinstance(key, tuple) and len(key) == 3:
            del self[key[0]][key[1]][key[2]]
        else:
            raise ValueError(f"Invalid key: {key}")


class PropertyOptions:
    """This is a descriptor class for instance properties in a typed class.

    It is used when you have a property that has a different name in the Data Model
    compared to the name in the Python class.

    Args:
        identifier (str | None): The name of the property in the Data Model. Defaults to the name of the property in the Python class.
    """

    def __init__(self, identifier: str | None = None) -> None:
        self.name = cast(str, identifier)  # mypy help, set_name guarantees str

    def __set_name__(self, owner: type, attribute_name: str) -> None:
        self.name = self.name or attribute_name
        if self.name in _RESERVED_PROPERTY_NAMES:
            self.name = _RESERVED_PROPERTY_CONFLICT_PREFIX + self.name

    def __get__(self, instance: TypedInstance, owner: type) -> Any:
        try:
            return instance.__dict__[self.name]
        except KeyError:
            raise AttributeError(f"'{owner.__name__}' object has no attribute '{self.name}'")

    def __set__(self, instance: TypedInstance, value: Any) -> None:
        try:
            instance.__dict__[self.name] = value
        except KeyError:
            raise AttributeError(f"'{instance.__class__.__name__}' object has no attribute '{self.name}'")

    def __delete__(self, instance: TypedInstance) -> None:
        try:
            del instance.__dict__[self.name]
        except KeyError:
            raise AttributeError(f"'{instance.__class__.__name__}' object has no attribute '{self.name}'")

    @property
    def property_name(self) -> str:
        return self.resolve_property(self.name)

    @staticmethod
    def resolve_property(name: str) -> str:
        if name.startswith(_RESERVED_PROPERTY_CONFLICT_PREFIX):
            return name[len(_RESERVED_PROPERTY_CONFLICT_PREFIX) :]
        return name


class TypedInstance(ABC):
    @classmethod
    @abstractmethod
    def get_source(cls) -> ViewId:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def _base_properties() -> set[str]:
        raise NotImplementedError

    def _dump_properties(self) -> dict[str, Any]:
        props = {key: prop for key, prop in vars(self).items() if key not in _RESERVED_PROPERTY_NAMES}
        return _PropertyValueSerializer.serialize_values(props, camel_case=True)

    @classmethod
    @lru_cache(32)
    def _get_constructor_parameters(cls) -> dict[str, inspect.Parameter]:
        return {key: param for key, param in inspect.signature(cls.__init__).parameters.items() if key != "self"}

    @classmethod
    def _load_instance(cls, resource: dict[str, Any], properties: dict[str, Any]) -> Self:
        return cls(**cls._load_base_properties(resource), **cls._load_properties(properties))

    @classmethod
    def _load_base_properties(cls, resource: dict[str, Any]) -> dict[str, Any]:
        args: dict[str, Any] = {}
        signature_params = cls._get_constructor_parameters()
        for key in cls._base_properties():
            camel_key = to_camel_case(key)
            if key in signature_params:
                param = signature_params[key]
                default = param.default
                if camel_key in resource:
                    args[key] = cls._deserialize_values(resource[camel_key], param)
                elif default is inspect.Parameter.empty:
                    args[key] = None
                else:
                    args[key] = cls._deserialize_values(default or None, param)
        return args

    @classmethod
    def _load_properties(cls, resource: dict[str, Any]) -> dict[str, Any]:
        output: dict[str, Any] = {}
        property_by_name = cls._get_properties_by_name()
        signature_params = cls._get_constructor_parameters()
        for name, parameter in signature_params.items():
            if name in cls._base_properties():
                continue
            if name in resource:
                output[name] = cls._deserialize_values(resource[name], parameter)
            elif name in property_by_name:
                prop = property_by_name[name].property_name
                if prop in resource:
                    output[name] = cls._deserialize_values(resource[prop], parameter)
        return output

    @classmethod
    @lru_cache(32)
    def _get_descriptor_property_name_mapping(cls) -> MappingProxyType[str, str]:
        return MappingProxyType(
            {
                v.property_name: k
                for base_class in cls.__mro__[:-1]
                for k, v in vars(base_class).items()
                if isinstance(v, PropertyOptions)
            }
        )

    @classmethod
    @lru_cache(32)
    def _get_user_defined_typed_instance_base_classes(cls) -> list[type]:
        # We want to exclude TypedInstance and its direct subclasses and all its superclasses.
        # So, only the base classes up to TypedNode, TypedNodeApply, etc...
        typed_instance_classes = {TypedInstance, *TypedInstance.__subclasses__(), *TypedInstance.mro()}
        return [base for base in cls.mro() if base not in typed_instance_classes]

    @classmethod
    def _get_properties_by_name(cls) -> dict[str, PropertyOptions]:
        property_by_name: dict[str, PropertyOptions] = {}
        for current_cls in cls._get_user_defined_typed_instance_base_classes():
            for name, value in vars(current_cls).items():
                if isinstance(value, PropertyOptions):
                    property_by_name[name] = value
        return property_by_name

    @classmethod
    def _deserialize_values(cls, value: Any, parameter: inspect.Parameter) -> Any:
        if isinstance(value, SequenceNotStr):
            # TODO, performance: Deserialize list of values should not inspect parameter/value repeatedly
            return [cls._deserialize_value(v, parameter) for v in value]
        else:
            return cls._deserialize_value(value, parameter)

    @staticmethod
    def _deserialize_value(value: Any, parameter: inspect.Parameter) -> Any:
        if parameter.annotation is inspect.Parameter.empty:
            return value
        annotation = str(parameter.annotation)
        if "datetime" in annotation and isinstance(value, str):
            return convert_data_modelling_timestamp(value)
        elif "date" in annotation and isinstance(value, str):
            return date.fromisoformat(value)
        elif isinstance(value, dict):
            if DirectRelationReference.__name__ in annotation:
                return DirectRelationReference.load(value)
            elif NodeId.__name__ in annotation:
                return NodeId.load(value)
            elif EdgeId.__name__ in annotation:
                return EdgeId.load(value)
        return value


class TypedNodeApply(NodeApply, TypedInstance):
    def __init__(
        self,
        space: str,
        external_id: str,
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
    ) -> None:
        super().__init__(space, external_id, existing_version, type=type)

    @staticmethod
    @lru_cache(1)
    def _base_properties() -> set[str]:
        return set(TypedNodeApply._get_constructor_parameters().keys())

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> Self:
        sources = resource.pop("sources", [])
        properties = sources[0].get("properties", {}) if sources else {}
        return cls._load_instance(resource, properties)

    @property
    def sources(self) -> list[NodeOrEdgeData]:
        return [NodeOrEdgeData(source=self.get_source(), properties=self._dump_properties())]


class TypedEdgeApply(EdgeApply, TypedInstance):
    def __init__(
        self,
        space: str,
        external_id: str,
        type: DirectRelationReference | tuple[str, str],
        start_node: DirectRelationReference | tuple[str, str],
        end_node: DirectRelationReference | tuple[str, str],
        existing_version: int | None = None,
    ) -> None:
        super().__init__(space, external_id, type, start_node, end_node, existing_version)

    @staticmethod
    @lru_cache(1)
    def _base_properties() -> set[str]:
        return set(TypedEdgeApply._get_constructor_parameters().keys())

    @classmethod
    def _load(cls, resource: dict, cognite_client: CogniteClient | None = None) -> Self:
        sources = resource.pop("sources", [])
        properties = sources[0].get("properties", {}) if sources else {}
        return cls._load_instance(resource, properties)

    @property
    def sources(self) -> list[NodeOrEdgeData]:
        return [NodeOrEdgeData(source=self.get_source(), properties=self._dump_properties())]


class TypedNode(Node, TypedInstance):
    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        deleted_time: int | None,
        type: DirectRelationReference | tuple[str, str] | None,
    ) -> None:
        super().__init__(
            space, external_id, version, last_updated_time, created_time, deleted_time, properties=None, type=type
        )

    @staticmethod
    @lru_cache(1)
    def _base_properties() -> set[str]:
        return set(TypedNode._get_constructor_parameters().keys())

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        all_properties = resource.pop("properties", {})
        source = cls.get_source()
        properties = all_properties.get(source.space, {}).get(source.as_source_identifier(), {})
        return cls._load_instance(resource, properties)

    @property
    def properties(self) -> Properties:
        return Properties({self.get_source(): self._dump_properties()})


class TypedEdge(Edge, TypedInstance):
    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        type: DirectRelationReference | tuple[str, str],
        last_updated_time: int,
        created_time: int,
        start_node: DirectRelationReference | tuple[str, str],
        end_node: DirectRelationReference | tuple[str, str],
        deleted_time: int | None,
    ) -> None:
        super().__init__(
            space,
            external_id,
            version,
            type,
            last_updated_time,
            created_time,
            start_node,
            end_node,
            deleted_time,
            properties=None,
        )

    @staticmethod
    @lru_cache(1)
    def _base_properties() -> set[str]:
        return set(TypedEdge._get_constructor_parameters().keys())

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        all_properties = resource.pop("properties", {})
        source = cls.get_source()
        properties = all_properties.get(source.space, {}).get(source.as_source_identifier(), {})
        return cls._load_instance(resource, properties)

    @property
    def properties(self) -> Properties:
        return Properties({self.get_source(): self._dump_properties()})


_RESERVED_PROPERTY_CONFLICT_PREFIX = "__________"
_RESERVED_PROPERTY_NAMES = (
    TypedNodeApply._base_properties()
    | TypedEdgeApply._base_properties()
    | TypedNode._base_properties()
    | TypedEdge._base_properties()
) | {"instance_type", "existing_version", "_InstanceApply__sources", "_Instance__properties", "_Instance__prop_lookup"}


class _PropertyValueSerializer:
    @classmethod
    def serialize_values(cls, props: Mapping[str, Any], camel_case: bool = True) -> dict[str, PropertyValue]:
        properties: dict[str, Any] = {}
        for key, value in props.items():
            key = PropertyOptions.resolve_property(key)
            if isinstance(value, SequenceNotStr):
                properties[key] = [cls._serialize_value(v, camel_case) for v in value]
            else:
                properties[key] = cls._serialize_value(value, camel_case)
        return properties

    @staticmethod
    def _serialize_value(value: Any, camel_case: bool) -> PropertyValue:
        if isinstance(value, NodeId):
            # We don't want to dump the instance_type field when serializing NodeId in this context
            return value.dump(camel_case, include_instance_type=False)
        elif isinstance(value, DirectRelationReference):
            return value.dump(camel_case)
        elif isinstance(value, datetime):
            return value.isoformat(timespec="milliseconds")
        elif isinstance(value, date):
            return value.isoformat()
        else:
            return value



from __future__ import annotations

from collections.abc import Iterator, Sequence
from typing import Literal, overload

from cognite.client._api_client import APIClient
from cognite.client._constants import DEFAULT_LIMIT_READ
from cognite.client.data_classes import (
    ThreeDModel,
    ThreeDModelList,
    ThreeDModelUpdate,
    ThreeDModelWrite,
)
from cognite.client.utils._identifier import IdentifierSequence, InternalId
from cognite.client.utils.useful_types import SequenceNotStr


class ThreeDModelsAPI(APIClient):
    _RESOURCE_PATH = "/3d/models"

    @overload
    def __call__(
        self, chunk_size: None = None, published: bool | None = None, limit: int | None = None
    ) -> Iterator[ThreeDModel]: ...

    @overload
    def __call__(
        self, chunk_size: int, published: bool | None = None, limit: int | None = None
    ) -> Iterator[ThreeDModelList]: ...

    def __call__(
        self, chunk_size: int | None = None, published: bool | None = None, limit: int | None = None
    ) -> Iterator[ThreeDModel] | Iterator[ThreeDModelList]:
        """Iterate over 3d models

        Fetches 3d models as they are iterated over, so you keep a limited number of 3d models in memory.

        Args:
            chunk_size (int | None): Number of 3d models to return in each chunk. Defaults to yielding one model a time.
            published (bool | None): Filter based on whether or not the model has published revisions.
            limit (int | None): Maximum number of 3d models to return. Defaults to return all items.

        Returns:
            Iterator[ThreeDModel] | Iterator[ThreeDModelList]: yields ThreeDModel one by one if chunk is not specified, else ThreeDModelList objects.
        """
        return self._list_generator(
            list_cls=ThreeDModelList,
            resource_cls=ThreeDModel,
            method="GET",
            chunk_size=chunk_size,
            filter={"published": published},
            limit=limit,
        )

    def __iter__(self) -> Iterator[ThreeDModel]:
        """Iterate over 3d models

        Fetches models as they are iterated over, so you keep a limited number of models in memory.

        Returns:
            Iterator[ThreeDModel]: yields models one by one.
        """
        return self()

    def retrieve(self, id: int) -> ThreeDModel | None:
        """`Retrieve a 3d model by id <https://developer.cognite.com/api#tag/3D-Models/operation/get3DModel>`_

        Args:
            id (int): Get the model with this id.

        Returns:
            ThreeDModel | None: The requested 3d model.

        Example:

            Get 3d model by id::

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.three_d.models.retrieve(id=1)
        """
        return self._retrieve(cls=ThreeDModel, identifier=InternalId(id))

    def list(self, published: bool | None = None, limit: int | None = DEFAULT_LIMIT_READ) -> ThreeDModelList:
        """`List 3d models. <https://developer.cognite.com/api#tag/3D-Models/operation/get3DModels>`_

        Args:
            published (bool | None): Filter based on whether or not the model has published revisions.
            limit (int | None): Maximum number of models to retrieve. Defaults to 25. Set to -1, float("inf") or None to return all items.

        Returns:
            ThreeDModelList: The list of 3d models.

        Examples:

            List 3d models:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> three_d_model_list = client.three_d.models.list()

            Iterate over 3d models:

                >>> for three_d_model in client.three_d.models:
                ...     three_d_model # do something with the 3d model

            Iterate over chunks of 3d models to reduce memory load:

                >>> for three_d_model in client.three_d.models(chunk_size=50):
                ...     three_d_model # do something with the 3d model
        """
        return self._list(
            list_cls=ThreeDModelList,
            resource_cls=ThreeDModel,
            method="GET",
            filter={"published": published},
            limit=limit,
        )

    def create(
        self,
        name: str | ThreeDModelWrite | SequenceNotStr[str | ThreeDModelWrite],
        data_set_id: int | None = None,
        metadata: dict[str, str] | None = None,
    ) -> ThreeDModel | ThreeDModelList:
        """`Create new 3d models. <https://developer.cognite.com/api#tag/3D-Models/operation/create3DModels>`_

        Args:
            name (str | ThreeDModelWrite | SequenceNotStr[str | ThreeDModelWrite]): The name of the 3d model(s) or 3D
                model object to create. If a 3D model object is provided, the other arguments are ignored.
            data_set_id (int | None): The id of the dataset this 3D model belongs to.
            metadata (dict[str, str] | None): Custom, application-specific metadata. String key -> String value.
                Limits: Maximum length of key is 32 bytes, value 512 bytes, up to 16 key-value pairs.

        Returns:
            ThreeDModel | ThreeDModelList: The created 3d model(s).

        Example:

            Create new 3d models::

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.three_d.models.create(name="My Model", data_set_id=1, metadata={"key1": "value1", "key2": "value2"})

            Create multiple new 3D Models::

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import ThreeDModelWrite
                >>> client = CogniteClient()
                >>> my_model = ThreeDModelWrite(name="My Model", data_set_id=1, metadata={"key1": "value1", "key2": "value2"})
                >>> my_other_model = ThreeDModelWrite(name="My Other Model", data_set_id=1, metadata={"key1": "value1", "key2": "value2"})
                >>> res = client.three_d.models.create([my_model, my_other_model])

        """
        items: ThreeDModelWrite | list[ThreeDModelWrite]
        if isinstance(name, str):
            items = ThreeDModelWrite(name, data_set_id, metadata)
        elif isinstance(name, ThreeDModelWrite):
            items = name
        else:
            items = [ThreeDModelWrite(n, data_set_id, metadata) if isinstance(n, str) else n for n in name]
        return self._create_multiple(list_cls=ThreeDModelList, resource_cls=ThreeDModel, items=items)

    @overload
    def update(
        self,
        item: ThreeDModel | ThreeDModelUpdate,
        mode: Literal["replace_ignore_null", "patch", "replace"] = "replace_ignore_null",
    ) -> ThreeDModel: ...

    @overload
    def update(
        self,
        item: Sequence[ThreeDModel | ThreeDModelUpdate],
        mode: Literal["replace_ignore_null", "patch", "replace"] = "replace_ignore_null",
    ) -> ThreeDModelList: ...

    def update(
        self,
        item: ThreeDModel | ThreeDModelUpdate | Sequence[ThreeDModel | ThreeDModelUpdate],
        mode: Literal["replace_ignore_null", "patch", "replace"] = "replace_ignore_null",
    ) -> ThreeDModel | ThreeDModelList:
        """`Update 3d models. <https://developer.cognite.com/api#tag/3D-Models/operation/update3DModels>`_

        Args:
            item (ThreeDModel | ThreeDModelUpdate | Sequence[ThreeDModel | ThreeDModelUpdate]): ThreeDModel(s) to update
            mode (Literal['replace_ignore_null', 'patch', 'replace']): How to update data when a non-update object is given (ThreeDModel or -Write). If you use 'replace_ignore_null', only the fields you have set will be used to replace existing (default). Using 'replace' will additionally clear all the fields that are not specified by you. Last option, 'patch', will update only the fields you have set and for container-like fields such as metadata or labels, add the values to the existing. For more details, see :ref:`appendix-update`.

        Returns:
            ThreeDModel | ThreeDModelList: Updated ThreeDModel(s)

        Examples:

            Update 3d model that you have fetched. This will perform a full update of the model:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> three_d_model = client.three_d.models.retrieve(id=1)
                >>> three_d_model.name = "New Name"
                >>> res = client.three_d.models.update(three_d_model)

            Perform a partial update on a 3d model:

                >>> from cognite.client.data_classes import ThreeDModelUpdate
                >>> my_update = ThreeDModelUpdate(id=1).name.set("New Name")
                >>> res = client.three_d.models.update(my_update)

        """
        # Note that we cannot use the ThreeDModelWrite to update as the write format of a 3D model
        # does not have ID or External ID, thus no identifier to know which model to update.
        return self._update_multiple(
            list_cls=ThreeDModelList,
            resource_cls=ThreeDModel,
            update_cls=ThreeDModelUpdate,
            items=item,
            mode=mode,
        )

    def delete(self, id: int | Sequence[int]) -> None:
        """`Delete 3d models. <https://developer.cognite.com/api#tag/3D-Models/operation/delete3DModels>`_

        Args:
            id (int | Sequence[int]): ID or list of IDs to delete.

        Example:

            Delete 3d model by id::

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.three_d.models.delete(id=1)
        """
        self._delete_multiple(identifiers=IdentifierSequence.load(ids=id), wrap_ids=True)



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Status(_message.Message):
    __slots__ = ("code", "symbol")
    CODE_FIELD_NUMBER: _ClassVar[int]
    SYMBOL_FIELD_NUMBER: _ClassVar[int]
    code: int
    symbol: str
    def __init__(self, code: _Optional[int] = ..., symbol: _Optional[str] = ...) -> None: ...

class NumericDatapoint(_message.Message):
    __slots__ = ("timestamp", "value", "status", "nullValue")
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    NULLVALUE_FIELD_NUMBER: _ClassVar[int]
    timestamp: int
    value: float
    status: Status
    nullValue: bool
    def __init__(self, timestamp: _Optional[int] = ..., value: _Optional[float] = ..., status: _Optional[_Union[Status, _Mapping]] = ..., nullValue: bool = ...) -> None: ...

class NumericDatapoints(_message.Message):
    __slots__ = ("datapoints",)
    DATAPOINTS_FIELD_NUMBER: _ClassVar[int]
    datapoints: _containers.RepeatedCompositeFieldContainer[NumericDatapoint]
    def __init__(self, datapoints: _Optional[_Iterable[_Union[NumericDatapoint, _Mapping]]] = ...) -> None: ...

class StringDatapoint(_message.Message):
    __slots__ = ("timestamp", "value", "status", "nullValue")
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    NULLVALUE_FIELD_NUMBER: _ClassVar[int]
    timestamp: int
    value: str
    status: Status
    nullValue: bool
    def __init__(self, timestamp: _Optional[int] = ..., value: _Optional[str] = ..., status: _Optional[_Union[Status, _Mapping]] = ..., nullValue: bool = ...) -> None: ...

class StringDatapoints(_message.Message):
    __slots__ = ("datapoints",)
    DATAPOINTS_FIELD_NUMBER: _ClassVar[int]
    datapoints: _containers.RepeatedCompositeFieldContainer[StringDatapoint]
    def __init__(self, datapoints: _Optional[_Iterable[_Union[StringDatapoint, _Mapping]]] = ...) -> None: ...

class AggregateDatapoint(_message.Message):
    __slots__ = ("timestamp", "average", "max", "min", "count", "sum", "interpolation", "stepInterpolation", "continuousVariance", "discreteVariance", "totalVariation", "countGood", "countUncertain", "countBad", "durationGood", "durationUncertain", "durationBad", "maxDatapoint", "minDatapoint")
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    AVERAGE_FIELD_NUMBER: _ClassVar[int]
    MAX_FIELD_NUMBER: _ClassVar[int]
    MIN_FIELD_NUMBER: _ClassVar[int]
    COUNT_FIELD_NUMBER: _ClassVar[int]
    SUM_FIELD_NUMBER: _ClassVar[int]
    INTERPOLATION_FIELD_NUMBER: _ClassVar[int]
    STEPINTERPOLATION_FIELD_NUMBER: _ClassVar[int]
    CONTINUOUSVARIANCE_FIELD_NUMBER: _ClassVar[int]
    DISCRETEVARIANCE_FIELD_NUMBER: _ClassVar[int]
    TOTALVARIATION_FIELD_NUMBER: _ClassVar[int]
    COUNTGOOD_FIELD_NUMBER: _ClassVar[int]
    COUNTUNCERTAIN_FIELD_NUMBER: _ClassVar[int]
    COUNTBAD_FIELD_NUMBER: _ClassVar[int]
    DURATIONGOOD_FIELD_NUMBER: _ClassVar[int]
    DURATIONUNCERTAIN_FIELD_NUMBER: _ClassVar[int]
    DURATIONBAD_FIELD_NUMBER: _ClassVar[int]
    MAXDATAPOINT_FIELD_NUMBER: _ClassVar[int]
    MINDATAPOINT_FIELD_NUMBER: _ClassVar[int]
    timestamp: int
    average: float
    max: float
    min: float
    count: float
    sum: float
    interpolation: float
    stepInterpolation: float
    continuousVariance: float
    discreteVariance: float
    totalVariation: float
    countGood: float
    countUncertain: float
    countBad: float
    durationGood: float
    durationUncertain: float
    durationBad: float
    maxDatapoint: NumericDatapoint
    minDatapoint: NumericDatapoint
    def __init__(self, timestamp: _Optional[int] = ..., average: _Optional[float] = ..., max: _Optional[float] = ..., min: _Optional[float] = ..., count: _Optional[float] = ..., sum: _Optional[float] = ..., interpolation: _Optional[float] = ..., stepInterpolation: _Optional[float] = ..., continuousVariance: _Optional[float] = ..., discreteVariance: _Optional[float] = ..., totalVariation: _Optional[float] = ..., countGood: _Optional[float] = ..., countUncertain: _Optional[float] = ..., countBad: _Optional[float] = ..., durationGood: _Optional[float] = ..., durationUncertain: _Optional[float] = ..., durationBad: _Optional[float] = ..., maxDatapoint: _Optional[_Union[NumericDatapoint, _Mapping]] = ..., minDatapoint: _Optional[_Union[NumericDatapoint, _Mapping]] = ...) -> None: ...

class AggregateDatapoints(_message.Message):
    __slots__ = ("datapoints",)
    DATAPOINTS_FIELD_NUMBER: _ClassVar[int]
    datapoints: _containers.RepeatedCompositeFieldContainer[AggregateDatapoint]
    def __init__(self, datapoints: _Optional[_Iterable[_Union[AggregateDatapoint, _Mapping]]] = ...) -> None: ...

class InstanceId(_message.Message):
    __slots__ = ("space", "externalId")
    SPACE_FIELD_NUMBER: _ClassVar[int]
    EXTERNALID_FIELD_NUMBER: _ClassVar[int]
    space: str
    externalId: str
    def __init__(self, space: _Optional[str] = ..., externalId: _Optional[str] = ...) -> None: ...



from __future__ import annotations

from cognite.client.data_classes.simulators.filters import (
    PropertySort,
    SimulationRunsSort,
)
from cognite.client.data_classes.simulators.logs import (
    SimulatorLog,
    SimulatorLogData,
    SimulatorLogList,
)
from cognite.client.data_classes.simulators.models import (
    SimulatorFlowsheet,
    SimulatorFlowsheetGraphicalObject,
    SimulatorFlowsheetObjectEdge,
    SimulatorFlowsheetObjectNode,
    SimulatorFlowsheetPosition,
    SimulatorFlowsheetProperty,
    SimulatorFlowsheetThermodynamic,
    SimulatorModel,
    SimulatorModelDependencyFileId,
    SimulatorModelDependencyFileReference,
    SimulatorModelList,
    SimulatorModelRevision,
    SimulatorModelRevisionData,
    SimulatorModelRevisionDataList,
    SimulatorModelRevisionDependency,
    SimulatorModelRevisionList,
    SimulatorModelRevisionWrite,
    SimulatorModelRevisionWriteList,
    SimulatorModelUpdate,
    SimulatorModelWrite,
    SimulatorModelWriteList,
)
from cognite.client.data_classes.simulators.routine_revisions import (
    SimulationValueUnitInput,
    SimulatorRoutineConfiguration,
    SimulatorRoutineDataSampling,
    SimulatorRoutineInput,
    SimulatorRoutineInputConstant,
    SimulatorRoutineInputList,
    SimulatorRoutineInputTimeseries,
    SimulatorRoutineLogicalCheck,
    SimulatorRoutineOutput,
    SimulatorRoutineOutputList,
    SimulatorRoutineRevision,
    SimulatorRoutineRevisionList,
    SimulatorRoutineRevisionWrite,
    SimulatorRoutineRevisionWriteList,
    SimulatorRoutineSchedule,
    SimulatorRoutineStage,
    SimulatorRoutineStageList,
    SimulatorRoutineSteadyStateDetection,
    SimulatorRoutineStep,
    SimulatorRoutineStepArguments,
)
from cognite.client.data_classes.simulators.routines import (
    SimulatorRoutine,
    SimulatorRoutineList,
    SimulatorRoutineWrite,
    SimulatorRoutineWriteList,
)
from cognite.client.data_classes.simulators.runs import (
    SimulationInput,
    SimulationInputOverride,
    SimulationOutput,
    SimulationRun,
    SimulationRunDataItem,
    SimulationRunDataList,
    SimulationRunList,
    SimulationRunWrite,
    SimulationRunWriteList,
    SimulationValueUnit,
    SimulationValueUnitName,
)
from cognite.client.data_classes.simulators.simulators import (
    Simulator,
    SimulatorIntegration,
    SimulatorIntegrationList,
    SimulatorList,
    SimulatorStep,
    SimulatorStepField,
    SimulatorStepOption,
    SimulatorUnitEntry,
)

__all__ = [
    "PropertySort",
    "SimulationInput",
    "SimulationInputOverride",
    "SimulationOutput",
    "SimulationRun",
    "SimulationRunDataItem",
    "SimulationRunDataList",
    "SimulationRunList",
    "SimulationRunWrite",
    "SimulationRunWriteList",
    "SimulationRunsSort",
    "SimulationValueUnit",
    "SimulationValueUnitInput",
    "SimulationValueUnitName",
    "Simulator",
    "SimulatorFlowsheet",
    "SimulatorFlowsheetGraphicalObject",
    "SimulatorFlowsheetObjectEdge",
    "SimulatorFlowsheetObjectNode",
    "SimulatorFlowsheetPosition",
    "SimulatorFlowsheetProperty",
    "SimulatorFlowsheetThermodynamic",
    "SimulatorIntegration",
    "SimulatorIntegrationList",
    "SimulatorList",
    "SimulatorLog",
    "SimulatorLogData",
    "SimulatorLogList",
    "SimulatorModel",
    "SimulatorModelDependencyFileId",
    "SimulatorModelDependencyFileReference",
    "SimulatorModelList",
    "SimulatorModelRevision",
    "SimulatorModelRevisionData",
    "SimulatorModelRevisionDataList",
    "SimulatorModelRevisionDependency",
    "SimulatorModelRevisionList",
    "SimulatorModelRevisionWrite",
    "SimulatorModelRevisionWriteList",
    "SimulatorModelUpdate",
    "SimulatorModelWrite",
    "SimulatorModelWriteList",
    "SimulatorRoutine",
    "SimulatorRoutineConfiguration",
    "SimulatorRoutineDataSampling",
    "SimulatorRoutineInput",
    "SimulatorRoutineInputConstant",
    "SimulatorRoutineInputList",
    "SimulatorRoutineInputTimeseries",
    "SimulatorRoutineList",
    "SimulatorRoutineLogicalCheck",
    "SimulatorRoutineOutput",
    "SimulatorRoutineOutputList",
    "SimulatorRoutineRevision",
    "SimulatorRoutineRevisionList",
    "SimulatorRoutineRevisionWrite",
    "SimulatorRoutineRevisionWriteList",
    "SimulatorRoutineSchedule",
    "SimulatorRoutineStage",
    "SimulatorRoutineStageList",
    "SimulatorRoutineSteadyStateDetection",
    "SimulatorRoutineStep",
    "SimulatorRoutineStepArguments",
    "SimulatorRoutineWrite",
    "SimulatorRoutineWriteList",
    "SimulatorStep",
    "SimulatorStepField",
    "SimulatorStepOption",
    "SimulatorUnitEntry",
]



from __future__ import annotations

from cognite.client._api_client import APIClient
from cognite.client.data_classes.units import (
    UnitSystem,
    UnitSystemList,
)


class UnitSystemAPI(APIClient):
    _RESOURCE_PATH = "/units/systems"

    def list(self) -> UnitSystemList:
        """`List all supported unit systems <https://developer.cognite.com/api#tag/Unit-Systems/operation/listUnitSystems>`_

        Returns:
            UnitSystemList: List of unit systems

        Examples:

            List all supported unit systems in CDF:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.units.systems.list()

        """
        return self._list(method="GET", list_cls=UnitSystemList, resource_cls=UnitSystem)



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Literal

from typing_extensions import Self

from cognite.client.data_classes._base import (
    CogniteObject,
    CogniteResource,
    CogniteResourceList,
    IdTransformerMixin,
)

if TYPE_CHECKING:
    from cognite.client import CogniteClient

Severity = Literal["Debug", "Information", "Warning", "Error"]


@dataclass
class SimulatorLogData(CogniteObject):
    """
    Simulator log data represents a single log entry in a simulator log.

    Args:
        timestamp (int): Timestamp of the log message.
        message (str): Log message.
        severity (Severity): Log severity level.
    """

    timestamp: int
    message: str
    severity: Severity

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            timestamp=resource["timestamp"],
            message=resource["message"],
            severity=resource["severity"],
        )


class SimulatorLog(CogniteResource):
    """
    Simulator logs track what happens during simulation runs, model parsing, and generic connector logic.
    They provide valuable information for monitoring, debugging, and auditing.

    Simulator logs capture important events, messages, and exceptions that occur during the execution of simulations, model parsing, and connector operations.
    They help users identify issues, diagnose problems, and gain insights into the behavior of the simulator integrations.

    Args:
        id (int): A unique id of a simulator resource log.
        data (Sequence[SimulatorLogData]): Log data of the simulator resource.
        created_time (int): The number of milliseconds since epoch.
        last_updated_time (int): The number of milliseconds since epoch.
        data_set_id (int): Dataset id of the resource.
        severity (Severity | None): Minimum severity level of the log data. This overrides connector configuration minimum severity level and can be used for more granular control.
    """

    def __init__(
        self,
        id: int,
        data: Sequence[SimulatorLogData],
        created_time: int,
        last_updated_time: int,
        data_set_id: int,
        severity: Severity | None,
    ) -> None:
        self.id = id
        self.created_time = created_time
        self.last_updated_time = last_updated_time
        self.data_set_id = data_set_id
        self.data = data
        self.severity = severity

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        return cls(
            id=resource["id"],
            created_time=resource["createdTime"],
            last_updated_time=resource["lastUpdatedTime"],
            data_set_id=resource["dataSetId"],
            data=[SimulatorLogData._load(entry, cognite_client) for entry in resource["data"]],
            severity=resource.get("severity"),
        )

    def dump(self, camel_case: bool = True) -> dict[str, Any]:
        output = super().dump(camel_case=camel_case)
        if isinstance(self.data, list) and all(isinstance(item, SimulatorLogData) for item in self.data):
            output["data"] = [item.dump(camel_case=camel_case) for item in self.data]

        return output


class SimulatorLogList(CogniteResourceList[SimulatorLog], IdTransformerMixin):
    _RESOURCE = SimulatorLog



from __future__ import annotations

from abc import ABC
from functools import cached_property
from urllib.parse import urljoin

from cognite.client._api_client import APIClient
from cognite.client.exceptions import CogniteAPIError


class OrgAPIClient(APIClient, ABC):
    _auth_url = "https://auth.cognite.com"

    def _get_base_url_with_base_path(self) -> str:
        """Get base URL with base path including organization and api version if applicable"""
        base_path = ""
        if self._api_version:
            base_path = f"/api/{self._api_version}/orgs/{self._organization}"
        # The OrganizationAPi uses the auth_url as the base for these endpoints instead of the
        # base_url like the rest of the SDK.
        return urljoin(self._auth_url, base_path)

    @cached_property
    def _organization(self) -> str:
        headers = self._configure_headers(
            "application/json",
            additional_headers=self._config.headers.copy(),
            api_subversion=self._api_subversion,
        )
        # This is an internal endpoint, not part of the public API
        full_url = urljoin(self._config.base_url, f"/api/v1/projects/{self._config.project}")
        response = self._http_client_with_retry.request(method="GET", url=full_url, headers=headers)
        if response.status_code != 200:
            raise CogniteAPIError(
                "Could not look-up organization", response.status_code, response.headers.get("x-request-id")
            )
        return response.json()["organization"]



from __future__ import annotations

from typing import TYPE_CHECKING

from cognite.client._api.raw.databases import RawDatabasesAPI
from cognite.client._api.raw.rows import RawRowsAPI
from cognite.client._api.raw.tables import RawTablesAPI
from cognite.client._api_client import APIClient

if TYPE_CHECKING:
    from cognite.client import CogniteClient
    from cognite.client.config import ClientConfig


class RawAPI(APIClient):
    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: CogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)
        self.databases = RawDatabasesAPI(config, api_version, cognite_client)
        self.tables = RawTablesAPI(config, api_version, cognite_client)
        self.rows = RawRowsAPI(config, api_version, cognite_client)
